import sys
import typing

import attrs
import pandas

import tecton_core.query.dialect
from tecton_core.compute_mode import ComputeMode
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_utils import get_pipeline_dialect
from tecton_core.query.nodes import AddAnchorTimeColumnsForSawtoothIntervalsNode
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import AddBooleanPartitionColumnsNode
from tecton_core.query.nodes import AddDurationNode
from tecton_core.query.nodes import AddEffectiveTimestampNode
from tecton_core.query.nodes import AddRetrievalAnchorTimeNode
from tecton_core.query.nodes import AddUniqueIdNode
from tecton_core.query.nodes import AdjustAnchorTimeToWindowEndNode
from tecton_core.query.nodes import AggregationSecondaryKeyExplodeNode
from tecton_core.query.nodes import AggregationSecondaryKeyRollupNode
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.query.nodes import AsofJoinInputContainer
from tecton_core.query.nodes import AsofJoinNode
from tecton_core.query.nodes import AsofJoinSawtoothAggNode
from tecton_core.query.nodes import AsofSecondaryKeyExplodeNode
from tecton_core.query.nodes import ConvertEpochToTimestampNode
from tecton_core.query.nodes import ConvertTimestampToUTCNode
from tecton_core.query.nodes import CustomFilterNode
from tecton_core.query.nodes import DataNode
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import DeriveValidityPeriodNode
from tecton_core.query.nodes import EntityFilterNode
from tecton_core.query.nodes import ExplodeEventsAndSelectDistinctNode
from tecton_core.query.nodes import ExplodeTimestampByTimeWindowsNode
from tecton_core.query.nodes import FeatureTimeFilterNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import InnerJoinOnRangeNode
from tecton_core.query.nodes import JoinNode
from tecton_core.query.nodes import MetricsCollectorNode
from tecton_core.query.nodes import MockDataSourceScanNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import OnlineListAggNode
from tecton_core.query.nodes import OnlinePartialAggNodeV2
from tecton_core.query.nodes import PartialAggNode
from tecton_core.query.nodes import PythonDataNode
from tecton_core.query.nodes import RawDataSourceScanNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import RespectFeatureStartTimeNode
from tecton_core.query.nodes import RespectTTLNode
from tecton_core.query.nodes import SelectDistinctNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import StreamWatermarkNode
from tecton_core.query.nodes import TakeLastRowNode
from tecton_core.query.nodes import TemporalBatchTableFormatNode
from tecton_core.query.nodes import TrimValidityPeriodNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.nodes import WildcardJoinNode
from tecton_spark.query import data_source
from tecton_spark.query import filter
from tecton_spark.query import join
from tecton_spark.query import pipeline
from tecton_spark.query import projection
from tecton_spark.query.node import SparkExecNode
from tecton_spark.schema_spark_utils import schema_to_spark


if sys.version_info >= (3, 9):
    from typing import get_args
    from typing import get_origin
else:
    from typing_extensions import get_args
    from typing_extensions import get_origin

from typing import List
from typing import Optional

import pyspark

from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_spark import data_source_helper


# NOTE: use repr=False to avoid printing out the underlying dataframes when used in REPL/notebook.
@attrs.define(repr=False)
class SparkDataFrame(DataframeWrapper):
    _dataframe: pyspark.sql.DataFrame

    @property
    def columns(self) -> List[str]:
        return self._dataframe.columns

    def to_pandas(self) -> pandas.DataFrame:
        return self._dataframe.toPandas()

    def to_spark(self) -> pyspark.sql.DataFrame:
        return self._dataframe


def _convert_snowflake_feature_views_to_spark_dataframes(qt_root: NodeRef, spark: pyspark.sql.SparkSession) -> None:
    """Finds all FeatureViewPipelineNodes, executes their subtrees, and replaces them with DataNodes.

    Assumes that all FeatureViewPipelineNodes have Snowflake SQL.
    """
    snowflake_fv_node_ref = get_first_input_node_of_class(qt_root, FeatureViewPipelineNode, as_node_ref=True)
    while snowflake_fv_node_ref is not None:
        data_node = _execute_snowflake_fv(snowflake_fv_node_ref, spark)
        snowflake_fv_node_ref.node = data_node
        snowflake_fv_node_ref = get_first_input_node_of_class(qt_root, FeatureViewPipelineNode, as_node_ref=True)


def _execute_snowflake_fv(snowflake_fv_node_ref: NodeRef, spark: pyspark.sql.SparkSession) -> DataNode:
    """Executes the FeatureViewPipelineNode and its subtree. Returns the results as a DataNode."""
    fv_node = snowflake_fv_node_ref.node
    if fv_node is None:
        msg = "The query tree must contain a FeatureViewPipelineNode with Snowflake SQL."
        raise RuntimeError(msg)

    # We compile the subtree starting at the FeatureViewPipelineNode into a SQL query and then execute it.
    fv_node = fv_node.with_dialect(Dialect.SNOWFLAKE)
    sql_string = fv_node.to_sql(pretty_sql=True)

    fv_spec = fv_node.feature_definition_wrapper.fv_spec
    spark_schema = schema_to_spark(fv_spec.view_schema)

    ds_node = get_first_input_node_of_class(snowflake_fv_node_ref, DataSourceScanNode)
    if ds_node is None:
        msg = "The query tree must contain a DataSourceScanNode with a Snowflake data source."
        raise RuntimeError(msg)
    snowflake_spec = ds_node.ds.batch_source

    snowflake_df = data_source_helper.get_snowflake_dataframe(
        spark,
        url=snowflake_spec.url,
        database=snowflake_spec.database,
        schema=snowflake_spec.schema,
        warehouse=snowflake_spec.warehouse,
        query=sql_string,
        spark_schema=spark_schema,
    )

    return DataNode(Dialect.SPARK, ComputeMode.SPARK, SparkDataFrame(snowflake_df))


# convert from logical tree to physical tree
def _spark_convert(node_ref: NodeRef) -> SparkExecNode:
    logical_tree_node = node_ref.node
    node_mapping = {
        CustomFilterNode: filter.CustomFilterSparkNode,
        DataSourceScanNode: data_source.DataSourceScanSparkNode,
        RawDataSourceScanNode: data_source.RawDataSourceScanSparkNode,
        MockDataSourceScanNode: data_source.MockDataSourceScanSparkNode,
        OfflineStoreScanNode: data_source.OfflineStoreScanSparkNode,
        FeatureViewPipelineNode: pipeline.PipelineEvalSparkNode,
        MultiOdfvPipelineNode: pipeline.MultiOdfvPipelineSparkNode,
        FeatureTimeFilterNode: filter.FeatureTimeFilterSparkNode,
        EntityFilterNode: filter.EntityFilterSparkNode,
        RespectTTLNode: filter.RespectTTLSparkNode,
        RespectFeatureStartTimeNode: filter.RespectFeatureStartTimeSparkNode,
        AddAnchorTimeNode: projection.AddAnchorTimeSparkNode,
        AddRetrievalAnchorTimeNode: projection.AddRetrievalAnchorTimeSparkNode,
        StreamWatermarkNode: filter.StreamWatermarkSparkNode,
        UserSpecifiedDataNode: data_source.UserSpecifiedDataSparkNode,
        DataNode: data_source.DataSparkNode,
        PartialAggNode: pipeline.PartialAggSparkNode,
        JoinNode: join.JoinSparkNode,
        WildcardJoinNode: join.WildcardJoinSparkNode,
        AsofJoinNode: join.AsofJoinSparkNode,
        AsofJoinFullAggNode: join.AsofJoinFullAggSparkNode,
        AsofSecondaryKeyExplodeNode: join.AsofSecondaryKeyExplodeSparkNode,
        RenameColsNode: projection.RenameColsSparkNode,
        SelectDistinctNode: projection.SelectDistinctSparkNode,
        ConvertEpochToTimestampNode: projection.ConvertEpochToTimestampSparkNode,
        ConvertTimestampToUTCNode: projection.ConvertTimestampToUTCSparkNode,
        AddEffectiveTimestampNode: projection.AddEffectiveTimestampSparkNode,
        MetricsCollectorNode: pipeline.MetricsCollectorSparkNode,
        AddDurationNode: projection.AddDurationSparkNode,
        AggregationSecondaryKeyRollupNode: join.AggregationSecondaryKeyRollupSparkNode,
        AggregationSecondaryKeyExplodeNode: projection.AggregationSecondaryKeyExplodeSparkNode,
        StagingNode: pipeline.StagingSparkNode,
        AddUniqueIdNode: projection.AddUniqueIdSparkNode,
        PythonDataNode: pipeline.PythonDataSparkNode,
        InnerJoinOnRangeNode: join.InnerJoinOnRangeSparkNode,
        OnlinePartialAggNodeV2: pipeline.OnlinePartialAggSparkNodeV2,
        OnlineListAggNode: pipeline.OnlineListAggSparkNode,
        TakeLastRowNode: join.TakeLastRowSparkNode,
        TemporalBatchTableFormatNode: join.TemporalBatchTableFormatSparkNode,
        ExplodeTimestampByTimeWindowsNode: join.ExplodeTimestampByTimeWindowsSparkNode,
        DeriveValidityPeriodNode: projection.DeriveValidityPeriodSparkNode,
        TrimValidityPeriodNode: filter.TrimValidityPeriodSparkNode,
        AddAnchorTimeColumnsForSawtoothIntervalsNode: projection.AddAnchorTimeColumnsForSawtoothIntervalsSparkNode,
        ExplodeEventsAndSelectDistinctNode: projection.ExplodeEventsAndSelectDistinctSparkNode,
        AddBooleanPartitionColumnsNode: projection.AddBooleanPartitionColumnsSparkNode,
        AdjustAnchorTimeToWindowEndNode: projection.AdjustAnchorTimeToWindowEndSparkNode,
        AsofJoinSawtoothAggNode: join.AsofJoinSawtoothAggSparkNode,
    }

    if logical_tree_node.__class__ in node_mapping:
        return node_mapping[logical_tree_node.__class__].from_query_node(logical_tree_node)

    msg = f"TODO: mapping for {logical_tree_node.__class__}"
    raise Exception(msg)


def spark_convert(qt_root: NodeRef, spark: Optional[pyspark.sql.SparkSession] = None) -> SparkExecNode:
    pipeline_dialect = get_pipeline_dialect(qt_root)
    if pipeline_dialect == tecton_core.query.dialect.Dialect.DUCKDB:
        msg = "Tecton on Spark does not support DuckDB."
        raise RuntimeError(msg)

    if pipeline_dialect == tecton_core.query.dialect.Dialect.SNOWFLAKE:
        if spark is None:
            msg = "A SparkSession is required in order to convert this query tree. It is required to eagerly execute the Snowflake transformations."
            raise RuntimeError(msg)
        _convert_snowflake_feature_views_to_spark_dataframes(qt_root, spark)

    return _spark_convert(qt_root)


def attrs_spark_converter(
    attrs_inst: typing.Any,  # noqa: ANN401
    attr: attrs.Attribute,
    item: typing.Any,  # noqa: ANN401
) -> typing.Any:  # noqa: ANN401
    """
    This converts a NodeRef into a SparkNode if it's a NodeRef.
    """
    # Check if type is a typing wrapper.
    if get_origin(attr.type) is not None:
        if get_origin(attr.type) == typing.Union and NodeRef in get_args(attr.type) and isinstance(item, NodeRef):
            return _spark_convert(item)
        elif get_origin(attr.type) == dict and NodeRef == get_args(attr.type)[1]:
            return {k: _spark_convert(v) for k, v in item.items()}
        return item
    if attr.type == NodeRef:
        return _spark_convert(item)
    if attr.type == AsofJoinInputContainer:
        return join.AsofJoinInputSparkContainer(
            **attrs.asdict(item, value_serializer=attrs_spark_converter, recurse=False)
        )
    return item



from typing import Sequence

from pyspark.sql import types as spark_types

from tecton import types as sdk_types
from tecton_core import data_types as tecton_types
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper


def to_column(field: sdk_types.Field) -> schema_pb2.Column:
    data_type = field.tecton_type().proto.data_type
    return schema_pb2.Column(name=field.name, offline_data_type=data_type)


def to_tecton_schema(fields: Sequence[sdk_types.Field]) -> schema_pb2.Schema:
    columns = [to_column(field) for field in fields]
    return schema_pb2.Schema(columns=columns)


def to_spark_schema_wrapper(field_list: Sequence[sdk_types.Field]) -> SparkSchemaWrapper:
    s = spark_types.StructType([field.spark_type() for field in field_list])
    return SparkSchemaWrapper(s)


def sdk_type_from_tecton_type(data_type: tecton_types.DataType) -> sdk_types.SdkDataType:
    if isinstance(data_type, tecton_types.Int32Type):
        return sdk_types.Int32
    elif isinstance(data_type, tecton_types.Int64Type):
        return sdk_types.Int64
    elif isinstance(data_type, tecton_types.Float32Type):
        return sdk_types.Float32
    elif isinstance(data_type, tecton_types.Float64Type):
        return sdk_types.Float64
    elif isinstance(data_type, tecton_types.StringType):
        return sdk_types.String
    elif isinstance(data_type, tecton_types.BoolType):
        return sdk_types.Bool
    elif isinstance(data_type, tecton_types.TimestampType):
        return sdk_types.Timestamp
    elif isinstance(data_type, tecton_types.ArrayType):
        return sdk_types.Array(sdk_type_from_tecton_type(data_type.element_type))
    elif isinstance(data_type, tecton_types.StructType):
        fields = [sdk_types.Field(field.name, sdk_type_from_tecton_type(field.data_type)) for field in data_type.fields]
        return sdk_types.Struct(fields)
    elif isinstance(data_type, tecton_types.MapType):
        return sdk_types.Map(
            sdk_type_from_tecton_type(data_type.key_type), sdk_type_from_tecton_type(data_type.value_type)
        )
    else:
        msg = f"{data_type} is not a recognized data types."
        raise NotImplementedError(msg)


def _schema_pretty_str_helper(schema, indent=0, schema_str=""):
    if isinstance(schema, Sequence):
        schema_str += (" " * indent) + "[\n"
        for item in schema:
            schema_str += _schema_pretty_str_helper(item, indent + 4)
        schema_str = schema_str[:-2] + "\n"  # remove trailing comma
        schema_str += " " * indent + "]\n"
    elif isinstance(schema, sdk_types.Field):
        schema_str += (" " * indent) + f"Field('{schema.name}', "
        schema_str += _schema_pretty_str_helper(schema.dtype, indent + 4)
        schema_str += "),\n"
    elif isinstance(schema, sdk_types.Struct):
        schema_str += "Struct(\n"
        schema_str += _schema_pretty_str_helper(schema.fields, indent + 4)
        schema_str += " " * indent + ")"
    elif isinstance(schema, sdk_types.Array):
        schema_str += "Array("
        schema_str += _schema_pretty_str_helper(schema.element_type, indent + 4)
        schema_str += ")"
    elif isinstance(schema, sdk_types.Map):
        schema_str += f"Map({_schema_pretty_str_helper(schema.key_type, indent)}, {_schema_pretty_str_helper(schema.value_type, indent)})"
    elif isinstance(schema, sdk_types.SdkDataType):
        schema_str += repr(schema)
    else:
        msg = f"Unknown type: {type(schema)}"
        raise ValueError(msg)

    return schema_str


def schema_pretty_str(schema: Sequence[sdk_types.Field]) -> str:
    """Returns a pretty, multi-line formatted string"""
    return _schema_pretty_str_helper(schema)[:-1]  # remove trailing new line character



import importlib
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple

import yaspin.spinners

from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli import repo_config as cli__repo_config
from tecton.cli.error_utils import pretty_error
from tecton.framework import base_tecton_object
from tecton.framework import repo_config as framework__repo_config
from tecton_core import conf
from tecton_core import repo_file_handler
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonValidationError


def _import_module_with_pretty_errors(
    file_path: Path,
    module_path: str,
    py_files: List[Path],
    repo_root: Path,
    before_error: Callable[[], None],
) -> ModuleType:
    from pyspark.sql.utils import AnalysisException

    try:
        module = importlib.import_module(module_path)
        if Path(module.__file__) != file_path:
            before_error()
            relpath = file_path.relative_to(repo_root)
            printer.safe_print(
                f"Python module name {cli_utils.bold(module_path)} ({relpath}) conflicts with module {module_path} from {module.__file__}. Please use a different name.",
                file=sys.stderr,
            )
            sys.exit(1)

        return module
    except AnalysisException as e:
        before_error()
        pretty_error(
            Path(file_path),
            py_files,
            exception=e,
            repo_root=repo_root,
            error_message="Analysis error",
            error_details=e.desc,
        )
        sys.exit(1)
    except TectonValidationError as e:
        before_error()
        pretty_error(Path(file_path), py_files, exception=e, repo_root=repo_root, error_message=e.args[0])
        sys.exit(1)
    except SyntaxError as e:
        before_error()
        details = None
        if e.text and e.offset:
            details = e.text + (" " * e.offset) + "^^^"
        pretty_error(
            Path(file_path),
            py_files,
            exception=e,
            repo_root=repo_root,
            error_message=e.args[0],
            error_details=details,
        )
        sys.exit(1)
    except TectonAPIInaccessibleError as e:
        before_error()
        printer.safe_print("Failed to connect to Tecton server at", e.args[1], ":", e.args[0])
        sys.exit(1)
    except Exception as e:
        before_error()
        pretty_error(Path(file_path), py_files, exception=e, repo_root=repo_root, error_message=e.args[0])
        sys.exit(1)


def collect_top_level_objects(
    py_files: List[Path], repo_root: Path, pretty_errors: bool
) -> List[base_tecton_object.BaseTectonObject]:
    modules = [cli_utils.py_path_to_module(p, repo_root) for p in py_files]

    with printer.safe_yaspin(yaspin.spinners.Spinners.earth, text="Importing feature repository modules") as sp:
        for file_path, module_path in zip(py_files, modules):
            sp.text = f"Processing feature repository module {module_path}"

            if pretty_errors:
                module = _import_module_with_pretty_errors(
                    file_path=file_path,
                    module_path=module_path,
                    py_files=py_files,
                    repo_root=repo_root,
                    before_error=lambda: sp.fail(printer.safe_string("⛔")),
                )
            else:
                module = importlib.import_module(module_path)

        num_modules = len(modules)
        sp.text = f"Imported {num_modules} Python {cli_utils.plural(num_modules, 'module', 'modules')} from the feature repository"
        sp.ok(printer.safe_string("✅"))

        return list(base_tecton_object._LOCAL_TECTON_OBJECTS)


def get_tecton_objects_skip_validation(
    specified_repo_config: Optional[Path],
) -> Tuple[List[base_tecton_object.BaseTectonObject], str, List[Path], Path]:
    repo_file_handler.ensure_prepare_repo()
    repo_files = repo_file_handler.repo_files()
    repo_root = repo_file_handler.repo_root()

    repo_config = _maybe_load_repo_config_or_default(specified_repo_config)

    py_files = [p for p in repo_files if p.suffix == ".py"]
    os.chdir(Path(repo_root))

    # Skip validate() during object initialization. We only do server-side validation in NDD environment
    with conf._temporary_set("TECTON_SKIP_OBJECT_VALIDATION", True), conf._temporary_set(
        "TECTON_REQUIRE_SCHEMA", False
    ):
        top_level_objects = collect_top_level_objects(py_files, repo_root=Path(repo_root), pretty_errors=True)

    return top_level_objects, repo_root, repo_files, repo_config


def _maybe_load_repo_config_or_default(repo_config_path: Optional[Path]) -> Path:
    if repo_config_path is None:
        repo_config_path = cli__repo_config.get_default_repo_config_path()

    if framework__repo_config.get_repo_config() is None:
        # Load the repo config. The repo config may have already been loaded if tecton objects were collected multiple
        # times, e.g. during `tecton plan` the objects are collected for tests and then the plan.
        cli__repo_config.load_repo_config(repo_config_path)

    return repo_config_path



import datetime
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import attrs
import pendulum

from tecton_core import feature_definition_wrapper
from tecton_core import query_consts
from tecton_core import schema
from tecton_core import time_utils
from tecton_core.specs import LifetimeWindowSpec
from tecton_core.specs import OnlineBatchTablePart as SpecOnlineBatchTablePart
from tecton_core.specs import OnlineBatchTablePartTile as SpecOnlineBatchTablePartTile
from tecton_core.specs import TimeWindowSpec
from tecton_core.specs import create_time_window_spec_from_data_proto
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2


@attrs.frozen
class AggregationGroup:
    """AggregationGroup represents a group of aggregate features to compute with a corresponding start/end.

    The typical usage of this will be in compaction jobs, where we will use the start/end time to determine
    eligible rows for each individual aggregate. The tile index and start/end time correspond to an OnlineBatchTablePartTile for this aggregation window group. Each tile may represent a smaller time window, within the larger aggregation window, to compute futures for.
    """

    window_index: int
    inclusive_start_time: Optional[datetime.datetime]
    exclusive_end_time: datetime.datetime
    aggregate_features: Tuple[feature_view__data_pb2.AggregateFeature, ...]
    schema: schema.Schema
    tile_index: int
    window_tile_column_name: str  # 0_0 for example


def _get_inclusive_start_time_for_window(
    exclusive_end_time: datetime.datetime,
    aggregation_window: TimeWindowSpec,
    tile_window: Optional[SpecOnlineBatchTablePartTile] = None,
) -> Optional[datetime.datetime]:
    if isinstance(aggregation_window, LifetimeWindowSpec):
        return None
    window_start_time = (
        tile_window.relative_start_time_inclusive if tile_window is not None else aggregation_window.window_start
    )
    return time_utils.get_timezone_aware_datetime(exclusive_end_time + window_start_time)


def _get_exclusive_end_time_for_window(
    exclusive_end_time: datetime.datetime,
    aggregation_window: TimeWindowSpec,
    tile_window: Optional[SpecOnlineBatchTablePartTile] = None,
) -> datetime.datetime:
    if isinstance(aggregation_window, LifetimeWindowSpec):
        return time_utils.get_timezone_aware_datetime(exclusive_end_time)
    window_end_time = (
        tile_window.relative_end_time_exclusive if tile_window is not None else aggregation_window.window_end
    )
    return time_utils.get_timezone_aware_datetime(exclusive_end_time + window_end_time)


def _get_groups_for_aggregation_part(
    agg_part: SpecOnlineBatchTablePart, exclusive_end_time: datetime.datetime, aggregation_map: Dict
) -> List[AggregationGroup]:
    """Create one aggregation group for each tile in each aggregation part.

    Stream FVs with non-lifetime windows use sawtooths and have multiple tiles per part.
    Batch feature views with non-lifetime windows have one tile set per part.
    Lifetime windows do not have tiles set in their OnlineBatchTablePart, but we still add 1 agg group per part.
    """
    tiles = agg_part.tiles or [None]
    agg_group_tiles = []
    for i, tile in enumerate(tiles):
        agg_group_tiles.append(
            AggregationGroup(
                window_index=agg_part.window_index,
                inclusive_start_time=_get_inclusive_start_time_for_window(
                    exclusive_end_time, agg_part.time_window, tile
                ),
                exclusive_end_time=_get_exclusive_end_time_for_window(exclusive_end_time, agg_part.time_window, tile),
                aggregate_features=tuple(aggregation_map[agg_part.time_window]),
                schema=agg_part.schema,
                tile_index=i,
                window_tile_column_name=f"{agg_part.window_index}_{i}",
            )
        )
    return agg_group_tiles


def aggregation_groups(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    exclusive_end_time: datetime.datetime,
) -> Tuple[AggregationGroup, ...]:
    aggregation_map = defaultdict(list)
    for aggregation in fdw.trailing_time_window_aggregation().features:
        aggregation_map[create_time_window_spec_from_data_proto(aggregation.time_window)].append(aggregation)

    aggregation_parts = fdw.fv_spec.online_batch_table_format.online_batch_table_parts

    if len(aggregation_parts) != len(aggregation_map):
        msg = "unexpected difference in length of the spec's online_batch_table_format and trailing_time_window_aggregation"
        raise ValueError(msg)

    agg_groups = []
    for agg_part in aggregation_parts:
        agg_groups += _get_groups_for_aggregation_part(agg_part, exclusive_end_time, aggregation_map)

    return tuple(agg_groups)


def _get_min_window_start_time(
    aggregation_groups: Tuple[AggregationGroup, ...], fdw: feature_definition_wrapper.FeatureDefinitionWrapper
) -> Optional[pendulum.DateTime]:
    contains_lifetime_agg = any(group.inclusive_start_time is None for group in aggregation_groups)
    if contains_lifetime_agg:
        return fdw.materialization_start_timestamp
    min_window_time = min(group.inclusive_start_time for group in aggregation_groups)
    return pendulum.instance(min_window_time)


def _get_max_window_end_time(aggregation_groups: Tuple[AggregationGroup, ...]) -> pendulum.DateTime:
    max_window_time = max(group.exclusive_end_time for group in aggregation_groups)
    return pendulum.instance(max_window_time)


def get_data_time_limits_for_compaction(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper, compaction_job_end_time: datetime.datetime
) -> Optional[pendulum.Period]:
    """Compute the time filter to be used for online compaction jobs.

    This determines how much data to read from the offline store.
    For aggregate fvs,
        start_time=earliest agg window start
        end_time=latest agg window end
    For non agg fvs,
        start_time=max(feature start time, compaction_job_end_time - ttl)
        end_time=compaction_job_end_time"""
    if fdw.materialization_start_timestamp is None:
        return None

    if fdw.is_temporal_aggregate:
        agg_groups = aggregation_groups(fdw=fdw, exclusive_end_time=compaction_job_end_time)
        start_time = _get_min_window_start_time(agg_groups, fdw)
        end_time = _get_max_window_end_time(agg_groups)
        return pendulum.Period(start_time, end_time)

    if not fdw.is_temporal:
        msg = "Expected fv to be of type temporal or temporal aggregate."
        raise Exception(msg)

    # respect ttl and feature start time for temporal fvs
    end_time = pendulum.instance(compaction_job_end_time)
    if fdw.serving_ttl:
        if not fdw.feature_start_timestamp:
            msg = "Expected feature start time to be set for temporal fvs when ttl is set."
            raise Exception(msg)
        job_time_minus_ttl = end_time - fdw.serving_ttl
        start_time = max(fdw.feature_start_timestamp, job_time_minus_ttl)
    elif fdw.feature_start_timestamp:
        start_time = fdw.feature_start_timestamp
    else:
        msg = "Expected ttl or feature start time to be set for temporal fvs."
        raise Exception(msg)
    return pendulum.Period(start_time, end_time)


def get_sorted_tile_column_names(agg_window_index: int, agg_groups: List[AggregationGroup]) -> List[str]:
    """Calculate the list of window_tile_column_names in ascending order for a given aggregation window index."""
    assert all(
        agg_window_index == agg_group.window_index for agg_group in agg_groups
    ), "All aggregation groups must have the same window index."
    expected_tile_indexes = set(range(len(agg_groups)))
    assert {
        agg_group.tile_index for agg_group in agg_groups
    } == expected_tile_indexes, "All aggregation groups must have unique tile indexes."

    sorted_tiles = sorted(
        agg_groups,
        key=lambda agg_group: agg_group.tile_index,
    )
    return [tile.window_tile_column_name for tile in sorted_tiles]


@attrs.frozen
class SawtoothAggregationData:
    """SawtoothAggregationData represents the batch sawtooth tiles for a Feature View.

    The typical usage of this will be in offline retrieval for Feature Views that use sawtooth aggregations (i.e. Stream Fvs with compaction enabled and non lifetime windows).
    """

    batch_sawtooth_tile_sizes: List[pendulum.Duration]

    # TODO(samantha): add support for stream tile size sawtooths.
    def contains_day_sawtooths(self) -> bool:
        return any(tile_size == pendulum.Duration(days=1) for tile_size in self.batch_sawtooth_tile_sizes)

    def contains_hour_sawtooths(self) -> bool:
        return any(tile_size == pendulum.Duration(hours=1) for tile_size in self.batch_sawtooth_tile_sizes)

    def get_unique_sawtooth_sizes(self) -> List[pendulum.Duration]:
        return list(set(self.batch_sawtooth_tile_sizes))

    def get_anchor_time_columns(self) -> List[str]:
        anchor_time_columns = []
        if self.contains_day_sawtooths():
            anchor_time_columns.append(query_consts.anchor_time_for_day_sawtooth())
        if self.contains_hour_sawtooths():
            anchor_time_columns.append(query_consts.anchor_time_for_hour_sawtooth())
        return anchor_time_columns

    def get_identifier_partition_columns(self) -> List[str]:
        partition_columns = []
        if self.contains_day_sawtooths():
            partition_columns.append(query_consts.is_day_sawtooth())
        if self.contains_hour_sawtooths():
            partition_columns.append(query_consts.is_hour_sawtooth())
        return partition_columns

    def get_truncated_anchor_time_partition_columns(self) -> List[str]:
        partition_columns = []
        if self.contains_day_sawtooths():
            partition_columns.append(query_consts.time_partition_column_for_date())
        if self.contains_hour_sawtooths():
            partition_columns.append(query_consts.time_partition_column_for_hour())
        return partition_columns

    def get_time_partition_column_for_sawtooth_size(self, sawtooth_size: pendulum.Duration) -> str:
        sawtooth_size_to_time_partition_map = {
            pendulum.Duration(days=1): query_consts.time_partition_column_for_date(),
            pendulum.Duration(hours=1): query_consts.time_partition_column_for_hour(),
        }

        if sawtooth_size not in sawtooth_size_to_time_partition_map:
            msg = f"Invalid sawtooth size: {sawtooth_size}"
            raise ValueError(msg)
        return sawtooth_size_to_time_partition_map[sawtooth_size]

    def get_anchor_time_column_for_sawtooth_size(self, sawtooth_size: pendulum.Duration) -> str:
        sawtooth_size_to_anchor_time_map = {
            pendulum.Duration(days=1): query_consts.anchor_time_for_day_sawtooth(),
            pendulum.Duration(hours=1): query_consts.anchor_time_for_hour_sawtooth(),
        }

        if sawtooth_size not in sawtooth_size_to_anchor_time_map:
            msg = f"Invalid sawtooth size: {sawtooth_size}"
            raise ValueError(msg)
        return sawtooth_size_to_anchor_time_map[sawtooth_size]

    def get_identifier_partition_column_for_sawtooth_size(self, sawtooth_size: pendulum.Duration) -> str:
        sawtooth_size_to_identifier_partition_map = {
            pendulum.Duration(days=1): query_consts.is_day_sawtooth(),
            pendulum.Duration(hours=1): query_consts.is_hour_sawtooth(),
        }

        if sawtooth_size not in sawtooth_size_to_identifier_partition_map:
            msg = f"Invalid sawtooth size: {sawtooth_size}"
            raise ValueError(msg)
        return sawtooth_size_to_identifier_partition_map[sawtooth_size]

    def get_anchor_time_to_partition_columns_map(self) -> Dict[str, str]:
        anchor_time_to_identifier_map = {}
        if self.contains_day_sawtooths():
            anchor_time_to_identifier_map[query_consts.anchor_time_for_day_sawtooth()] = query_consts.is_day_sawtooth()
        if self.contains_hour_sawtooths():
            anchor_time_to_identifier_map[query_consts.anchor_time_for_hour_sawtooth()] = (
                query_consts.is_hour_sawtooth()
            )
        return anchor_time_to_identifier_map

    def get_anchor_time_to_timedelta_map(self, use_zero_timedelta: bool = False) -> Dict[str, datetime.timedelta]:
        anchor_time_to_timedelta_map = {}
        if self.contains_day_sawtooths():
            day_duration = datetime.timedelta(seconds=0) if use_zero_timedelta else datetime.timedelta(days=1)
            anchor_time_to_timedelta_map[query_consts.anchor_time_for_day_sawtooth()] = day_duration
        if self.contains_hour_sawtooths():
            hour_duration = datetime.timedelta(seconds=0) if use_zero_timedelta else datetime.timedelta(hours=1)
            anchor_time_to_timedelta_map[query_consts.anchor_time_for_hour_sawtooth()] = hour_duration
        return anchor_time_to_timedelta_map



import typing
from typing import Union

import attrs
import pandas

from tecton_core import conf
from tecton_core.query import node_interface
from tecton_core.query import node_utils
from tecton_core.query import nodes
from tecton_core.query.pandas import node
from tecton_core.query.pandas import nodes as pandas_nodes
from tecton_core.query.pandas import sql
from tecton_snowflake.query import nodes as snowflake_nodes
from tecton_snowflake.query import rewrite as snowflake_rewrite


if typing.TYPE_CHECKING:
    import snowflake.snowpark


@attrs.frozen
class SnowparkExecutor(sql.SqlExecutor):
    session: "snowflake.snowpark.Session"

    def read_sql(self, sql_string: str) -> pandas.DataFrame:
        pandas_df = self.session.sql(sql_string).toPandas()
        return pandas_df

    def sql_to_snowpark(self, sql_string: str) -> "snowflake.snowpark.DataFrame":
        return self.session.sql(sql_string)


def snowflake_convert(
    node_ref: node_interface.NodeRef, sql_executor: SnowparkExecutor, pretty_sql: bool
) -> Union[node.SqlExecNode, node.PandasExecNode, snowflake_nodes.SnowparkExecNode]:
    snowflake_rewrite.SnowflakeRewrite().rewrite(node_ref)

    if node_utils.tree_contains(node_ref, nodes.MultiOdfvPipelineNode):
        # conf to configure running odfvs using the Pandas exec mode instead of snowpark
        if conf.get_bool("SNOWFLAKE_PANDAS_ODFV_ENABLED"):
            return _convert_to_pandas_nodes(node_ref, sql_executor, pretty_sql)
        else:
            return _convert_to_snowpark_nodes(node_ref, sql_executor, pretty_sql)

    return node.SqlExecNode.from_sql_inputs(node_ref.node, sql_executor, pretty_sql)


def _convert_to_pandas_nodes(
    tree: node_interface.NodeRef, sql_executor: SnowparkExecutor, pretty_sql: bool
) -> Union[node.PandasExecNode, node.SqlExecNode]:
    # Recurses over RenameColsNodes and MultiOdfvPipelineNode at the top of the tree and converts them to
    # PandasExecNodes and converts only the node immediately below these nodes to a SqlExecNode
    logical_tree_node = tree.node
    node_mapping = {
        nodes.MultiOdfvPipelineNode: pandas_nodes.PandasMultiOdfvPipelineNode,
        nodes.RenameColsNode: pandas_nodes.PandasRenameColsNode,
    }
    if logical_tree_node.__class__ in node_mapping:
        input_node = _convert_to_pandas_nodes(logical_tree_node.input_node, sql_executor, pretty_sql)
        return node_mapping[logical_tree_node.__class__].from_node_inputs(
            logical_tree_node, input_node, column_name_updater=lambda x: x.upper()
        )
    else:
        return node.SqlExecNode.from_sql_inputs(logical_tree_node, sql_executor, pretty_sql)


def _convert_to_snowpark_nodes(
    tree: node_interface.NodeRef, sql_executor: SnowparkExecutor, pretty_sql: bool
) -> [node.SqlExecNode, snowflake_nodes.SnowparkExecNode]:
    # Recurses over RenameColsNodes and MultiOdfvPipelineNode at the top of the tree and converts them to
    # SnowparkExecNodes and converts only the node immediately below these nodes to a SqlExecNode
    logical_tree_node = tree.node
    node_mapping = {
        nodes.MultiOdfvPipelineNode: snowflake_nodes.SnowparkMultiOdfvPipelineNode,
        nodes.RenameColsNode: snowflake_nodes.SnowparkRenameColsNode,
    }
    if logical_tree_node.__class__ in node_mapping:
        input_node = _convert_to_snowpark_nodes(logical_tree_node.input_node, sql_executor, pretty_sql)
        return node_mapping[logical_tree_node.__class__].from_node_inputs(
            logical_tree_node, input_node, column_name_updater=lambda x: x.upper(), session=sql_executor.session
        )
    else:
        return node.SqlExecNode.from_sql_inputs(logical_tree_node, sql_executor, pretty_sql)



from dataclasses import dataclass
from typing import Callable
from typing import List

import pypika
from pypika import analytics
from pypika import functions
from pypika import terms

from tecton_core.aggregation_utils import get_materialization_aggregation_column_prefixes
from tecton_core.query import aggregation_plans
from tecton_proto.common import aggregation_function__client_pb2 as afpb
from tecton_snowflake.query import queries


@dataclass
class AggregationPlan:
    """
    An AggregationPlan contains all the methods required to compute feature values for a specific Tecton aggregation.
    Snowflake does not support RANGE BETWEEN on sliding window, so all the full aggregation terms and some of the partial aggregation terms are different from the tecton_core implementation.

    The order of the columns must be the same in:
    * the return list in partial_aggregation_query
    * the arguments list in full_aggregation_query
    * materialized_column_prefixes

    Attributes:
        full_aggregation_join_query_term: A method that maps a list of input partial aggregate columns and a QueryWindowSpec to an output pypika column containing the full aggregates.
        materialized_column_prefixes: The list of prefixes that should be applied to the columns produced by `partial_aggregation_transform`.
    """

    partial_aggregation_query_terms: Callable[[str], List[terms.Term]]
    full_aggregation_join_query_term: Callable[[List[str]], terms.Term]
    materialized_column_prefixes: List[str]

    def materialized_column_names(self, input_column_name: str) -> List[str]:
        return [f"{prefix}_{input_column_name}" for prefix in self.materialized_column_prefixes]


def get_aggregation_plan(
    aggregation_function: afpb.AggregationFunction,
    function_params: afpb.AggregationFunctionParams,
    time_key: str,
) -> AggregationPlan:
    plan = _SNOWFLAKE_AGGREGATION_PLANS.get(aggregation_function)
    if plan is None:
        msg = f"Unsupported aggregation function {aggregation_function}"
        raise ValueError(msg)

    if callable(plan):
        return plan(time_key, function_params)
    else:
        return plan


def _default_partial_aggregation_query_terms(
    function: afpb.AggregationFunction, input_column_name: str
) -> List[terms.Term]:
    return aggregation_plans.AGGREGATION_PLANS.get(function).partial_aggregation_query_terms(input_column_name)


def _simple_aggregation_plan(
    aggregation_function: afpb.AggregationFunction,
    row_function: terms.AggregateFunction,
):
    return AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(aggregation_function, col),
        full_aggregation_join_query_term=lambda cols: row_function(pypika.Field(cols[0])),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(aggregation_function),
    )


def _mean_full_aggregation_join(cols: List[str]):
    mean_col, count_col = cols
    sum_query_term = analytics.Sum(pypika.Field(mean_col) * pypika.Field(count_col))
    count_query_term = analytics.Sum(pypika.Field(count_col))
    return sum_query_term / count_query_term


# sample variation equation: (Σ(x^2) - (Σ(x)^2)/N)/N-1
def _var_samp_full_aggregation_join(cols: List[str]):
    sum_of_squares_col, count_col, sum_col = cols
    count_query_term = functions.Cast(analytics.Sum(pypika.Field(count_col)), "double")
    sum_of_squares_query_term = functions.Cast(analytics.Sum(pypika.Field(sum_of_squares_col)), "double")
    sum_query_term = functions.Cast(analytics.Sum(pypika.Field(sum_col)), "double")
    # check if count is equal to 0 for divide by 0 errors
    var_samp_col = (sum_of_squares_query_term - (sum_query_term**2) / count_query_term) / functions.NullIf(
        count_query_term - 1, 0
    )
    return var_samp_col


def _var_pop_full_aggregation_join(cols: List[str]):
    sum_of_squares_col, count_col, sum_col = cols
    count_query_term = functions.Cast(analytics.Sum(pypika.Field(count_col)), "double")
    sum_of_squares_query_term = functions.Cast(analytics.Sum(pypika.Field(sum_of_squares_col)), "double")
    sum_query_term = functions.Cast(analytics.Sum(pypika.Field(sum_col)), "double")
    return (sum_of_squares_query_term / count_query_term) - (sum_query_term / count_query_term) ** 2


def _stddev_samp_full_aggregation_join(cols: List[str]):
    return functions.Sqrt(_var_samp_full_aggregation_join(cols))


def _stddev_pop_full_aggregation_join(cols: List[str]):
    return functions.Sqrt(_var_pop_full_aggregation_join(cols))


def _mean_aggregation_plan():
    return AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_MEAN, col
        ),
        full_aggregation_join_query_term=lambda cols: _mean_full_aggregation_join(cols),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_MEAN),
    )


def _make_last_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def last_non_distinct_n_partial_aggregation(col: str) -> List[terms.Term]:
        # Sort items in ascending order based on timestamp.
        return [queries.ArraySlice(queries.ArrayAggWithinGroup(col, time_key), 0, n)]

    return last_non_distinct_n_partial_aggregation


def _make_first_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def first_non_distinct_n_partial_aggregation(col: str) -> List[terms.Term]:
        # Sort items in ascending order based on timestamp.
        return [
            queries.ArraySlice(queries.ArrayAggWithinGroup(col, time_key), -n, queries.ArraySize(queries.ArrayAgg(col)))
        ]

    return first_non_distinct_n_partial_aggregation


def _make_last_non_distinct_n_full_aggregation(time_key: str, n: int) -> Callable:
    def _last_non_distinct_n_full_aggregation(cols: List[str]) -> terms.Term:
        return queries.ArraySlice(
            queries.ArrayAggWithinGroup("VALUE", f"{time_key}, INDEX"), -n, queries.ArraySize(queries.ArrayAgg(cols[0]))
        )

    return _last_non_distinct_n_full_aggregation


def _make_first_non_distinct_n_full_aggregation(time_key: str, n: int) -> Callable:
    def _first_non_distinct_n_full_aggregation(cols: List[str]) -> terms.Term:
        return queries.ArraySlice(queries.ArrayAggWithinGroup("VALUE", f"{time_key}, INDEX"), 0, n)

    return _first_non_distinct_n_full_aggregation


_SNOWFLAKE_AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_SUM, functions.Sum),
    afpb.AGGREGATION_FUNCTION_MIN: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MIN, functions.Min),
    afpb.AGGREGATION_FUNCTION_MAX: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MAX, functions.Max),
    afpb.AGGREGATION_FUNCTION_COUNT: AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_COUNT, col
        ),
        full_aggregation_join_query_term=lambda cols: functions.Sum(pypika.Field(cols[0])),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_COUNT),
    ),
    afpb.AGGREGATION_FUNCTION_MEAN: _mean_aggregation_plan(),
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_VAR_SAMP, col
        ),
        full_aggregation_join_query_term=lambda cols: _var_samp_full_aggregation_join(cols),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_VAR_SAMP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_VAR_POP: AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_VAR_POP, col
        ),
        full_aggregation_join_query_term=lambda cols: _var_pop_full_aggregation_join(cols),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(afpb.AGGREGATION_FUNCTION_VAR_POP),
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_STDDEV_SAMP, col
        ),
        full_aggregation_join_query_term=lambda cols: _stddev_samp_full_aggregation_join(cols),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_STDDEV_SAMP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: AggregationPlan(
        partial_aggregation_query_terms=lambda col: _default_partial_aggregation_query_terms(
            afpb.AGGREGATION_FUNCTION_STDDEV_POP, col
        ),
        full_aggregation_join_query_term=lambda cols: _stddev_pop_full_aggregation_join(cols),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_STDDEV_POP
        ),
    ),
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: lambda time_key, params: AggregationPlan(
        partial_aggregation_query_terms=_make_first_non_distinct_n_partial_aggregation(time_key, params.first_n.n),
        full_aggregation_join_query_term=_make_first_non_distinct_n_full_aggregation(time_key, params.first_n.n),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N, function_params=params
        ),
    ),
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: lambda time_key, params: AggregationPlan(
        partial_aggregation_query_terms=_make_last_non_distinct_n_partial_aggregation(time_key, params.last_n.n),
        full_aggregation_join_query_term=_make_last_non_distinct_n_full_aggregation(time_key, params.last_n.n),
        materialized_column_prefixes=get_materialization_aggregation_column_prefixes(
            afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N, function_params=params
        ),
    ),
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/workspace__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)tecton_proto/data/workspace__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a,tecton_proto/auditlog/metadata__client.proto\"\xa5\x01\n\x15WorkspaceCapabilities\x12/\n\x0ematerializable\x18\x01 \x01(\x08\x42\x07\xea\x99\xbd\x46\x02\x08\x02R\x0ematerializable\x12[\n\"offline_store_subdirectory_enabled\x18\x02 \x01(\x08:\x05\x66\x61lseB\x07\xea\x99\xbd\x46\x02\x08\x01R\x1fofflineStoreSubdirectoryEnabled\"\xa8\x01\n\tWorkspace\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12L\n\x0c\x63\x61pabilities\x18\x02 \x01(\x0b\x32(.tecton_proto.data.WorkspaceCapabilitiesR\x0c\x63\x61pabilities\x12\x39\n\ncreated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAtB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.workspace__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _WORKSPACECAPABILITIES.fields_by_name['materializable']._options = None
  _WORKSPACECAPABILITIES.fields_by_name['materializable']._serialized_options = b'\352\231\275F\002\010\002'
  _WORKSPACECAPABILITIES.fields_by_name['offline_store_subdirectory_enabled']._options = None
  _WORKSPACECAPABILITIES.fields_by_name['offline_store_subdirectory_enabled']._serialized_options = b'\352\231\275F\002\010\001'
  _WORKSPACECAPABILITIES._serialized_start=144
  _WORKSPACECAPABILITIES._serialized_end=309
  _WORKSPACE._serialized_start=312
  _WORKSPACE._serialized_end=480
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from dataclasses import field
from typing import Callable
from typing import Dict
from typing import List
from typing import Union

import pandas
import pyarrow

from tecton_core import data_types
from tecton_core.errors import TectonValidationError
from tecton_core.schema import Schema


@dataclass
class FieldTypeDiff:
    field: str
    msg: str


@dataclass
class DiffResult:
    missing_fields: List[str] = field(default_factory=list)
    missmatch_types: List[FieldTypeDiff] = field(default_factory=list)

    def __bool__(self):
        return len(self.missing_fields) > 0 or len(self.missmatch_types) > 0

    def as_str(self):
        def lines():
            if self.missing_fields:
                yield f"Missing fields: {', '.join(self.missing_fields)}"

            if self.missmatch_types:
                yield "Types do not match:"
                for item in self.missmatch_types:
                    yield f"  Field {item.field}: {item.msg}"

        return "\n".join(lines())


class CastError(TectonValidationError):
    def __init__(self, msg):
        super().__init__(msg, can_drop_traceback=True)

    @staticmethod
    def for_diff(diff_result: DiffResult) -> "CastError":
        return CastError("Schema mismatch:\n" + diff_result.as_str())


_ColumnGetter = Callable[[str, pyarrow.DataType], pyarrow.Array]


def cast(
    obj: Union[pyarrow.Table, pyarrow.RecordBatch, pandas.DataFrame], schema: Union[Schema, pyarrow.Schema]
) -> pyarrow.Table:
    if isinstance(schema, Schema):
        arrow_schema = tecton_schema_to_arrow_schema(schema)
    elif isinstance(schema, pyarrow.Schema):
        arrow_schema = schema
    else:
        msg = f"Unsupported schema type: {type(schema)}"
        raise TypeError(msg)

    if isinstance(obj, (pyarrow.RecordBatch, pyarrow.Table)):

        def get_column(name: str, dtype: pyarrow.DataType) -> pyarrow.Array:
            return obj.column(name).cast(dtype)

    elif isinstance(obj, pandas.DataFrame):
        columns = _pandas_columns(obj)

        def get_column(name: str, dtype: pyarrow.DataType) -> pyarrow.Array:
            series = columns[name]
            if len(series) != 1:
                msg = f"Ambiguous column label {name}. Ensure only one column exists with a given label."
                raise CastError(msg)
            return pyarrow.Array.from_pandas(series[0], type=dtype)

    else:
        msg = f"Unexpected type: {type(obj)}"
        raise TypeError(msg)
    arrays = cast_columns(get_column, arrow_schema)
    return pyarrow.Table.from_arrays(arrays, schema=arrow_schema)


def cast_batch(batch: pyarrow.RecordBatch, schema: Union[Schema, pyarrow.Schema]) -> pyarrow.RecordBatch:
    if isinstance(schema, Schema):
        arrow_schema = tecton_schema_to_arrow_schema(schema)
    elif isinstance(schema, pyarrow.Schema):
        arrow_schema = schema
    else:
        msg = f"Unsupported schema type: {type(schema)}"
        raise TypeError(msg)

    def get_column(name: str, dtype: pyarrow.DataType) -> pyarrow.Array:
        return batch.column(name).cast(dtype)

    arrays = cast_columns(get_column, arrow_schema)
    return pyarrow.RecordBatch.from_arrays(arrays, schema=arrow_schema)


def _pandas_columns(df: pandas.DataFrame) -> Dict[str, List[pandas.Series]]:
    def _series_iter():
        axes = df.axes
        if len(axes) != 2:
            msg = f"Pandas DataFrame should have 2 axes; not {len(axes)}"
            raise CastError(msg)

        index = df.index
        if isinstance(index, pandas.MultiIndex):
            for level_name in index.names:
                yield level_name, index.get_level_values(level_name)
        elif isinstance(index, pandas.Index):
            if index.name is not None:
                yield index.name, index
        else:
            msg = "First axis of a Pandas DataFrame should be an Index"
            raise CastError(msg)

        yield from df.items()

    ret = {}
    for label, series in _series_iter():
        ret.setdefault(label, []).append(series)
    return ret


def cast_columns(column_getter: _ColumnGetter, schema: pyarrow.Schema) -> List[pyarrow.Array]:
    diff = DiffResult()
    arrays = []
    for name, dtype in zip(schema.names, schema.types):
        try:
            arrays.append(column_getter(name, dtype))
        except KeyError:
            diff.missing_fields.append(name)
        except pyarrow.ArrowTypeError as e:
            diff.missmatch_types.append(FieldTypeDiff(name, str(e)))
        except pyarrow.ArrowInvalid as e:
            diff.missmatch_types.append(FieldTypeDiff(name, str(e)))
    if diff:
        raise CastError.for_diff(diff)
    else:
        return arrays


_PRIMITIVE_TECTON_TYPE_TO_ARROW_TYPE: Dict[data_types.DataType, pyarrow.DataType] = {
    data_types.Int32Type(): pyarrow.int32(),
    data_types.Int64Type(): pyarrow.int64(),
    data_types.Float32Type(): pyarrow.float32(),
    data_types.Float64Type(): pyarrow.float64(),
    data_types.StringType(): pyarrow.string(),
    data_types.TimestampType(): pyarrow.timestamp("ns", "UTC"),
    data_types.BoolType(): pyarrow.bool_(),
}


def _tecton_type_to_arrow_type(tecton_type: data_types.DataType) -> pyarrow.DataType:
    if tecton_type in _PRIMITIVE_TECTON_TYPE_TO_ARROW_TYPE:
        return _PRIMITIVE_TECTON_TYPE_TO_ARROW_TYPE[tecton_type]

    if isinstance(tecton_type, data_types.ArrayType):
        return pyarrow.list_(_tecton_type_to_arrow_type(tecton_type.element_type))

    if isinstance(tecton_type, data_types.MapType):
        return pyarrow.map_(
            _tecton_type_to_arrow_type(tecton_type.key_type),
            _tecton_type_to_arrow_type(tecton_type.value_type),
        )

    if isinstance(tecton_type, data_types.StructType):
        fields = []
        for tecton_field in tecton_type.fields:
            fields.append(pyarrow.field(tecton_field.name, _tecton_type_to_arrow_type(tecton_field.data_type)))
        return pyarrow.struct(fields)

    msg = f"Tecton type {tecton_type} can't be converted to arrow type"
    raise ValueError(msg)


def tecton_schema_to_arrow_schema(schema: Schema) -> pyarrow.Schema:
    fields = []
    for column, data_type in schema.column_name_and_data_types():
        fields.append(pyarrow.field(column, _tecton_type_to_arrow_type(data_type)))
    return pyarrow.schema(fields)



import random
import string

import sqlparse


def generate_random_name() -> str:
    return "".join(random.choice(string.ascii_uppercase) for i in range(20))


def format_sql(sql_str: str) -> str:
    return sqlparse.format(sql_str, reindent=True)



import datetime
import logging
import time
from typing import List

import pendulum
from google.protobuf.timestamp_pb2 import Timestamp
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.functions import col
from pyspark.sql.functions import explode
from pyspark.sql.functions import expr
from pyspark.sql.functions import lit
from pyspark.sql.functions import row_number
from pyspark.sql.types import LongType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType
from pyspark.sql.utils import AnalysisException
from pyspark.sql.window import Window

from tecton_core import specs
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.id_helper import IdHelper
from tecton_core.offline_store import OfflineStoreType
from tecton_core.offline_store import get_offline_store_partition_params
from tecton_core.offline_store import get_offline_store_type
from tecton_core.query_consts import anchor_time
from tecton_core.time_utils import convert_timestamp_for_version
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.common.task_params import job_query_intervals
from tecton_materialization.consumption import export_consumption_debug_metrics
from tecton_materialization.consumption import export_consumption_metrics
from tecton_materialization.job_metadata import is_checkpoint_complete
from tecton_materialization.job_metadata import write_checkpoint
from tecton_materialization.materialization_utils import MaterializationResourceContainer
from tecton_materialization.materialization_utils import batch_write_to_online_store
from tecton_materialization.materialization_utils import df_to_online_store_msg
from tecton_materialization.materialization_utils import has_prior_delta_commit
from tecton_materialization.materialization_utils import wait_for_metric_scrape
from tecton_proto.data.feature_view__client_pb2 import DeltaOfflineStoreVersion
from tecton_proto.materialization.job_metadata__client_pb2 import OnlineStoreType
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark import materialization_plan
from tecton_spark.data_observability import create_feature_metrics_collector
from tecton_spark.materialization_plan import MATERIALIZED_RAW_DATA_END_TIME
from tecton_spark.materialization_plan import MaterializationPlan
from tecton_spark.offline_store import OfflineStoreWriterParams
from tecton_spark.offline_store import get_offline_store_reader
from tecton_spark.offline_store import get_offline_store_writer
from tecton_spark.time_utils import convert_timestamp_to_epoch


logger = logging.getLogger(__name__)

# Partitions being written to S3 should be small to minimize number of parquet files.
COALESCE_FOR_SMALL_PARTITIONS = 1
DEFAULT_COALESCE_FOR_S3 = 10
# Partitions being written to OSW can be higher to better utilize available executors.
DEFAULT_COALESCE_FOR_OSW = 64

idempotence_key = "featureStartTime"


def _construct_anchor_times(
    fdw: FeatureDefinitionWrapper, start_time: datetime.datetime, num_tiles: int, version: int
) -> List[int]:
    """Creates `num_tiles` consecutive anchor_times starting from `start_time`.

    :return: An increasing list of consecutive anchor times.
    """
    anchor_times = []
    for i in range(num_tiles):
        anchor_time = start_time + i * fdw.get_tile_interval
        anchor_time_val = convert_timestamp_to_epoch(anchor_time, version)
        anchor_times.append(anchor_time_val)

    return anchor_times


def _construct_tile_end_times(
    fdw: FeatureDefinitionWrapper, latest_tile_end_time: datetime.datetime, num_tiles: int, version: int
) -> List[int]:
    """Creates `num_tiles` consecutive tile_end_times where latest one ends at `latest_tile_end_time`.

    :return: An increasing list of consecutive tile end times.
    """
    tile_end_times = []
    for i in range(num_tiles):
        tile_end_time = latest_tile_end_time - i * fdw.batch_materialization_schedule
        time_val = convert_timestamp_to_epoch(tile_end_time, version)
        tile_end_times.append(time_val)

    tile_end_times.reverse()
    return tile_end_times


def _dedupe_online_store_writes(
    fd: FeatureDefinitionWrapper,
    df: DataFrame,
) -> DataFrame:
    # this mimics the conditional writes in the OnlineStoreWriter
    if fd.is_temporal:
        # we take the latest record for each entity for temporal FVs
        window = Window.partitionBy(fd.join_keys).orderBy(col(fd.time_key).desc())
        row_number_col = "__tecton_row_num"
        assert row_number_col not in df.columns
        df = df.withColumn(row_number_col, row_number().over(window))
        df = df.filter(col(row_number_col) == 1).drop(row_number_col)
        return df
    else:
        return df


def _make_dynamodb_json_dataframe(
    spark: SparkSession,
    fd: FeatureDefinitionWrapper,
    materialization_task_params: MaterializationTaskParams,
    df: DataFrame,
) -> DataFrame:
    df = df_to_online_store_msg(
        df,
        fd.id,
        is_batch=True,
        is_status=False,
        canary_id=None,
        is_compaction_job=fd.compaction_enabled,
    )
    udf_name = f"to_dynamodb_json_{fd.id}"
    if fd.compaction_enabled:
        # UDF used for compaction jobs. Compatible with TableFormatVersionV3
        spark._jvm.com.tecton.onlinestorewriter.RegisterFeatureToDynamoDbJsonUDFV2().register(
            udf_name, materialization_task_params.SerializeToString()
        )
        udf_column_name = "dynamodb_json"
        df = df.select(expr(f"{udf_name}(value) as {udf_column_name}"))
        return df.select(explode(udf_column_name))
    else:
        # UDF used by bulk backfill jobs for non-compaction fvs.
        spark._jvm.com.tecton.onlinestorewriter.RegisterFeatureToDynamoDbJsonUDF().register(
            udf_name, materialization_task_params.SerializeToString()
        )
        return df.select(expr(f"{udf_name}(value)"))


def _write_dynamodb_json(
    spark: SparkSession,
    fd: FeatureDefinitionWrapper,
    materialization_task_params: MaterializationTaskParams,
    df: DataFrame,
) -> int:
    dynamodb_json_df = _make_dynamodb_json_dataframe(spark, fd, materialization_task_params, df)
    output_path = materialization_task_params.batch_task_info.dynamodb_json_output_path
    # we repartition the rows to avoid hot partitions (as recommended by AWS for dynamodb:importtable) and generate
    # enough output files to get good parallelism on the dynamo side
    # 128 is a magic number that should work well for large files but not too much overhead for small files; but it
    # is merely theoretical, feel free to change
    dynamodb_json_df.repartition(128).write.format("text").option("compression", "gzip").mode("overwrite").save(
        output_path
    )
    return dynamodb_json_df.count()


def _make_batch_status_dataframe(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    fd: FeatureDefinitionWrapper,
    feature_start_time: datetime.datetime,
    feature_end_time: datetime.datetime,
) -> DataFrame:
    # For continuous aggregation when writing uncompacted data we will default to writing
    # tile end times
    version = fd.get_feature_store_format_version
    batch_params = materialization_task_params.batch_task_info.batch_parameters
    if fd.is_temporal_aggregate and not fd.is_continuous:
        # For BWAFVs, write all materialized anchor times in the status table
        anchor_times = _construct_anchor_times(fd, feature_start_time, batch_params.tile_count, version)
        anchor_times_df_format = [[x] for x in anchor_times]
        schema = StructType([StructField(anchor_time(), LongType())])
        return spark.createDataFrame(anchor_times_df_format, schema=schema)
    else:
        # For BFVs, write materialized tile end times in the status table
        tile_end_times = _construct_tile_end_times(fd, feature_end_time, batch_params.tile_count, version)
        tile_end_times_df_format = [[x] for x in tile_end_times]
        schema = StructType([StructField(MATERIALIZED_RAW_DATA_END_TIME, LongType())])
        return spark.createDataFrame(tile_end_times_df_format, schema=schema)


def _write_batch_to_hdfs(plan: MaterializationPlan, hdfs_output_path: str):
    plan.base_data_frame.write.parquet(hdfs_output_path)


def _get_time_column_for_offline(fd: FeatureDefinitionWrapper) -> str:
    if (
        get_offline_store_type(fd) == OfflineStoreType.DELTA
        and fd.offline_store_params is not None
        and fd.offline_store_params.delta.version == DeltaOfflineStoreVersion.DELTA_OFFLINE_STORE_VERSION_2
    ):
        return fd.time_key

    if fd.is_temporal_aggregate:
        return anchor_time()

    return fd.time_key


def _materialize_batch_to_offline(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    feature_start_time: datetime.datetime,
    feature_end_time: datetime.datetime,
    plan: MaterializationPlan,
    job_metadata_client: JobMetadataClient,
):
    fd = feature_definition_from_task_params(materialization_task_params)
    version = fd.get_feature_store_format_version
    is_delta_offline_store = get_offline_store_type(fd) == OfflineStoreType.DELTA
    coalesce = (
        COALESCE_FOR_SMALL_PARTITIONS
        if get_offline_store_partition_params(fd).partition_interval.as_timedelta() <= datetime.timedelta(hours=1)
        else DEFAULT_COALESCE_FOR_S3
    )
    batch_task_info = materialization_task_params.batch_task_info
    should_avoid_coalesce = batch_task_info.should_avoid_coalesce
    offline_store_df = (
        plan.offline_store_data_frame if should_avoid_coalesce else plan.offline_store_data_frame.coalesce(coalesce)
    )

    offline_store_params = OfflineStoreWriterParams(
        s3_path=materialization_task_params.offline_store_path,
        always_store_anchor_column=True,
        time_column=_get_time_column_for_offline(fd),
        join_key_columns=fd.join_keys,
        is_continuous=fd.is_continuous,
    )

    logger.info(f"Writing to offline store for FV {fd.id} for interval {feature_start_time} to {feature_end_time}")
    start = time.time()

    store_writer = get_offline_store_writer(offline_store_params, fd, version, spark)

    start_time_proto = Timestamp()
    start_time_proto.FromDatetime(feature_start_time)
    # Need to replace previous the time range if doing a manual retry of a past job or if prev attempt wrote data
    if is_delta_offline_store and (
        has_prior_delta_commit(spark, materialization_task_params, idempotence_key, start_time_proto.ToJsonString())
        or materialization_task_params.batch_task_info.batch_parameters.is_overwrite
    ):
        tile = pendulum.period(feature_start_time, feature_end_time)
        store_writer.overwrite_dataframe_in_tile(offline_store_df, tile)
    else:
        store_writer.append_dataframe(offline_store_df)

    store_reader = get_offline_store_reader(spark, fd, path=materialization_task_params.offline_store_path)
    unhandled_analysis_exception = False
    try:
        start_time_col = convert_timestamp_for_version(feature_start_time, version)
        end_time_col = convert_timestamp_for_version(feature_end_time, version)
        # We need to subtract 1 microsecond from the end time to ensure that we don't
        # read the next tile
        offline_store_count = (
            store_reader.read(
                pendulum.period(feature_start_time, feature_end_time - datetime.timedelta(microseconds=1))
            )
            .filter((col(anchor_time()) >= lit(start_time_col)) & (col(anchor_time()) < lit(end_time_col)))
            .count()
        )
    except AnalysisException as e:
        is_df_empty = len(offline_store_df.take(1)) == 0
        if not is_df_empty:
            unhandled_analysis_exception = True
            logger.info(f"Unhandled AnalysisException checking Dataframe size: {e}")
        offline_store_count = 0
    export_consumption_debug_metrics(spark, offline_store_count, plan.cached_count(), unhandled_analysis_exception)

    latency = time.time() - start
    logger.info(f"Finished writing to offline store ({latency}s)")

    export_consumption_metrics(
        spark,
        fd,
        plan.cached_count(),
        materialization_task_params,
        store_type="offline",
        job_metadata_client=job_metadata_client,
    )


def _materialize_batch_to_online(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    feature_start_time: datetime.datetime,
    feature_end_time: datetime.datetime,
    plan: MaterializationPlan,
    resources,
    job_metadata_client: JobMetadataClient,
):
    batch_task_info = materialization_task_params.batch_task_info
    batch_params = batch_task_info.batch_parameters
    fv_spec = specs.create_feature_view_spec_from_data_proto(materialization_task_params.feature_view)
    fd = feature_definition_from_task_params(materialization_task_params)
    online_store_df = plan.online_store_data_frame
    start = time.time()

    write_type_detail = "bulk load intermediate storage" if batch_params.create_online_table else "online store"
    logger.info(
        f"Writing to {write_type_detail} for FV {fd.id} for interval {feature_start_time} to {feature_end_time}"
    )
    if batch_task_info.should_dedupe_online_store_writes:
        online_store_df = _dedupe_online_store_writes(fd, online_store_df)

    if batch_params.create_online_table:
        # note that it's okay to write the status table entries here ahead of time, because it won't be
        # marked as servable until the import table is complete
        # additionally, the status table range never shrinks, so even if we retry, it will be okay
        dynamo_import_row_count = _write_dynamodb_json(
            spark,
            fd,
            materialization_task_params,
            online_store_df,
        )
        # we export consumption metrics here for dynamo import instead of OSW because we don't use OSW
        export_consumption_metrics(
            spark,
            fd,
            dynamo_import_row_count,
            materialization_task_params,
            store_type="online",
            online_store_type=OnlineStoreType.ONLINE_STORE_TYPE_DYNAMO,
            job_metadata_client=job_metadata_client,
        )
        logger.info(f"Wrote {dynamo_import_row_count} rows to bulk load intermediate storage")
    else:
        assert not fv_spec.compaction_enabled, "Batch compaction must use create_online_table"
        # Write materialized features to the online feature store
        online_store_df = online_store_df.coalesce(DEFAULT_COALESCE_FOR_OSW)
        batch_write_to_online_store(
            online_store_df, materialization_task_params, resources.get_online_store_sink(), fd.id, is_status=False
        )
        export_consumption_metrics(
            spark,
            fd,
            plan.cached_count(),
            materialization_task_params,
            store_type="online",
            job_metadata_client=job_metadata_client,
            online_store_type=OnlineStoreType.ONLINE_STORE_TYPE_DYNAMO,
            only_new_consumption=True,
        )
    status_df = _make_batch_status_dataframe(
        spark, materialization_task_params, fd, feature_start_time, feature_end_time
    )

    batch_write_to_online_store(
        status_df, materialization_task_params, resources.get_online_store_sink(), fd.id, is_status=True
    )
    latency = time.time() - start
    logger.info(f"Finished writing to {write_type_detail} ({latency}s)")


def _get_hdfs_path(materialization_task_params: MaterializationTaskParams, end_time):
    hdfs_path_base = f"transformed_data_{IdHelper.to_string(materialization_task_params.attempt_id)}"
    formatted_time = end_time.strftime("%Y_%m_%d_%H_%M_%S")
    return f"{hdfs_path_base}_{formatted_time}"


def batch_materialize_from_params(
    spark: SparkSession,
    materialization_task_params: MaterializationTaskParams,
    run_id,
    job_metadata_client: JobMetadataClient,
    step=None,
):
    batch_params = materialization_task_params.batch_task_info.batch_parameters
    resources = MaterializationResourceContainer(spark, materialization_task_params)
    spark.conf.set(
        "spark.databricks.delta.commitInfo.userMetadata",
        f'{{"{idempotence_key}":"{batch_params.feature_start_time.ToJsonString()}"}}',
    )
    fd = feature_definition_from_task_params(materialization_task_params)
    task_feature_start_time = batch_params.feature_start_time.ToDatetime()
    task_feature_end_time = batch_params.feature_end_time.ToDatetime()

    # we only use checkpoints in steps that write to actual stores and have multiple intervals to write
    use_checkpoints = step != 1 and not materialization_task_params.HasField("canary_id") and fd.is_incremental_backfill

    write_to_offline_store_from_source = batch_params.write_to_offline_feature_store
    # if we are reading from the offline store, we don't need to split up the query, like we do from source
    # we actually don't want to, because dynamo json will not support this
    write_to_online_from_source = (
        batch_params.write_to_online_feature_store and not batch_params.read_from_offline_store_for_online_write
    )
    write_to_online_from_offline_store = (
        batch_params.write_to_online_feature_store and batch_params.read_from_offline_store_for_online_write
    )

    reads_data_from_source = write_to_offline_store_from_source or write_to_online_from_source
    if step == 1 and not reads_data_from_source:
        # we only need step 1 if we are reading data from source (which executes transformations)
        logger.info("Skipping step 1 since no data is read from source")
        return

    if write_to_offline_store_from_source or write_to_online_from_source:
        total_rows = 0
        # when reading from source, we must split the queries up into intervals based on incremental_backfills
        # note that we don't need to do this for 2 step materialization, but we keep it here to preserve old behavior
        # and not break existing customer jobs
        for query_feature_start_time, query_feature_end_time in job_query_intervals(materialization_task_params):
            if use_checkpoints and is_checkpoint_complete(spark, materialization_task_params, query_feature_start_time):
                logger.info(
                    "Checkpoint complete for interval %s to %s; skipping",
                    query_feature_start_time,
                    query_feature_end_time,
                )
                continue
            query_feature_time_limits = pendulum.instance(query_feature_end_time) - pendulum.instance(
                query_feature_start_time
            )
            if step in (1, None):
                metrics_collector = create_feature_metrics_collector(
                    spark,
                    materialization_task_params,
                    feature_start_time=query_feature_start_time,
                    feature_end_time=query_feature_end_time,
                )
                plan = materialization_plan.get_batch_materialization_plan(
                    spark=spark,
                    feature_definition=fd,
                    feature_data_time_limits=query_feature_time_limits,
                    metrics_collector=metrics_collector,
                )
                # we must cache here so metrics_collector does not double-count rows. However, this is not
                # as the cache is subject to eviction
                # TODO(dataobs): fix this
                plan.base_data_frame.cache()
            else:
                assert step == 2
                # we don't collect metrics in step 2, since they are already collected in step 1
                metrics_collector = None
                # for step 2, we read from the transformed data written in step 1
                plan = MaterializationPlan.from_parquet(
                    spark=spark, fd=fd, path=_get_hdfs_path(materialization_task_params, query_feature_end_time)
                )
            if step == 1:
                # for step 1, we write the transformed data to hdfs, to be read in step 2
                _write_batch_to_hdfs(plan, _get_hdfs_path(materialization_task_params, query_feature_end_time))
            else:
                if write_to_offline_store_from_source:
                    _materialize_batch_to_offline(
                        spark,
                        materialization_task_params,
                        query_feature_start_time,
                        query_feature_end_time,
                        plan,
                        job_metadata_client=job_metadata_client,
                    )
                if write_to_online_from_source:
                    _materialize_batch_to_online(
                        spark,
                        materialization_task_params,
                        query_feature_start_time,
                        query_feature_end_time,
                        plan,
                        resources,
                        job_metadata_client=job_metadata_client,
                    )
            if metrics_collector:
                try:
                    metrics_collector.publish()
                except Exception as e:
                    logger.error(f"Metrics publishing failed: {e}")
            if step in (None, 2):
                logger.info(
                    f"Wrote {plan.cached_count()} rows for interval {query_feature_start_time} to {query_feature_end_time}"
                )
                total_rows += plan.cached_count()
            if use_checkpoints:
                write_checkpoint(spark, materialization_task_params, query_feature_start_time, run_id)

        logger.info(f"Wrote {total_rows} total rows")

    if write_to_online_from_offline_store and step in (None, 2):
        assert not materialization_task_params.HasField("canary_id")
        # we don't do metrics collection here, because they already got collected in the jobs writing to the offline store
        if fd.compaction_enabled:
            plan = materialization_plan.get_batch_compaction_online_materialization_plan(
                spark=spark, feature_definition=fd, compaction_job_end_time=task_feature_end_time
            )
        else:
            # used for bootstrap bulk backfill jobs
            plan = MaterializationPlan.from_offline_store(fd, task_feature_start_time, task_feature_end_time, spark)
        if plan is not None:
            _materialize_batch_to_online(
                spark,
                materialization_task_params,
                task_feature_start_time,
                task_feature_end_time,
                plan,
                resources,
                job_metadata_client=job_metadata_client,
            )
        else:
            logger.info("Offline store empty; skipping write to online store")

    wait_for_metric_scrape()
    resources.close()



import urllib
from abc import ABC
from abc import abstractmethod
from typing import Optional
from urllib.parse import ParseResult

import attrs


class PrefixedUriResolver(ABC):
    @abstractmethod
    def uri(self, key: str) -> ParseResult:
        pass


@attrs.frozen
class PrefixedS3ResolverImpl(PrefixedUriResolver):
    bucket: Optional[str]

    def uri(self, key: str) -> ParseResult:
        if not self.bucket:
            msg = "Need to specify an S3 bucket to write "
            raise Exception(msg)
        return urllib.parse.urlparse(f"s3://{self.bucket}/{key}")


@attrs.frozen
class PrefixedLocalFileResolverImpl(PrefixedUriResolver):
    base_dir: str

    def uri(self, key: str) -> ParseResult:
        return urllib.parse.urlparse(f"file://{self.base_dir}/{key}")



import os
import platform
import re
import string

from yaspin import yaspin
from yaspin.spinners import Spinners


"""
A drop-in wrapper around `print` and `yaspin` that's primarily used to filter rich output
(i.e. emojis) from CLI output.

By default, rich output is disabled in non-Mac environments or if TECTON_RICH_OUTPUT != "1".
"""


def safe_string(s: str) -> str:
    if not _rich_output():
        return _filter_nonprintable(_filter_colors(_convert_emoji(s)))
    return s


def safe_print(*objects, **kwargs):
    # Mirrors Python3 print API (https://docs.python.org/3/library/functions.html#print)
    filtered_objects = []
    for o in objects:
        if isinstance(o, str):
            o = safe_string(o)
        filtered_objects.append(o)
    print(*filtered_objects, **kwargs)


def safe_yaspin(spinner, text):
    if not _rich_output():
        return yaspin(Spinners.simpleDots, text=text)
    return yaspin(spinner, text)


def _rich_output() -> bool:
    return platform.system() != "Windows" and (os.environ.get("TECTON_RICH_OUTPUT", "1") == "1")


def _filter_colors(s) -> str:
    # https://stackoverflow.com/questions/30425105/filter-special-chars-such-as-color-codes-from-shell-output
    return re.sub(r"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))", "", s)


def _filter_nonprintable(s) -> str:
    printable = set(string.printable)
    return "".join(filter(lambda x: x in printable, s))


def _convert_emoji(s) -> str:
    emoji_replacements = {
        "⛔": "[FAIL]",
        "✅": "[OK]",
        "💡": "[Tip]",
        "🎉": "[OK]",
        "⚠️": "[WARNING]",
        "↓": "-",
        "↑": "-",
        "⏳": "[IN_PROGRESS]",
        "🔎": "[DEBUG]",
    }
    for k, v in emoji_replacements.items():
        s = s.replace(k, v)
    return s



import inspect
from typing import Optional
from typing import Union

import attrs
import pandas
import pyarrow

from tecton_core import conf
from tecton_core.query.dialect import Dialect
from tecton_core.query.duckdb.compute import DuckDBCompute
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.pandas.nodes import PandasDataSourceScanNode
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.query.query_tree_compute import logger
from tecton_core.schema import Schema
from tecton_core.secrets import SecretResolver
from tecton_core.specs import FileSourceSpec


@attrs.frozen
class PandasCompute(SQLCompute):
    # For executing pipelines, Pandas will execute only the data source scan + pipeline nodes. Other
    # logic e.g. around asof joins are executed using DuckDB.
    sql_compute: DuckDBCompute

    @staticmethod
    def from_context() -> "PandasCompute":
        return PandasCompute(sql_compute=DuckDBCompute.from_context())

    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        return self.sql_compute.run_sql(
            sql_string,
            return_dataframe,
            expected_output_schema=expected_output_schema,
            monitor=monitor,
        )

    def get_dialect(self) -> Dialect:
        return Dialect.DUCKDB

    def register_temp_table_from_pandas(self, table_name: str, pandas_df: pandas.DataFrame) -> None:
        self.sql_compute.register_temp_table_from_pandas(table_name, pandas_df)

    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        self.sql_compute.register_temp_table(table_name, table_or_reader)

    def register_temp_table_from_data_source(
        self,
        table_name: str,
        ds: DataSourceScanNode,
        secret_resolver: Optional[SecretResolver] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> None:
        if isinstance(ds.ds.batch_source, FileSourceSpec):
            return self.sql_compute.register_temp_table_from_data_source(table_name, ds, secret_resolver, monitor)

        pandas_node = PandasDataSourceScanNode.from_node_inputs(
            query_node=ds, input_node=None, secret_resolver=secret_resolver
        )
        try:
            monitor.set_query(inspect.getsource(ds.ds.batch_source.function))
        except OSError:
            pass

        self.register_temp_table_from_pandas(table_name, pandas_node.to_dataframe())

    def load_table(self, table_name: str, expected_output_schema: Optional[Schema] = None) -> pyarrow.Table:
        return self.sql_compute.load_table(table_name, expected_output_schema=expected_output_schema)

    def run_odfv(
        self, qt_node: NodeRef, input: pyarrow.RecordBatchReader, monitor: Optional[ComputeMonitor] = None
    ) -> pyarrow.RecordBatchReader:
        from tecton_core.query.pandas.translate import pandas_convert_odfv_only

        if conf.get_bool("DUCKDB_DEBUG"):
            logger.warning(f"Input dataframe to ODFV execution: {input.schema}")

        pandas_node = pandas_convert_odfv_only(qt_node, input)
        # ToDo: extract code from pandas_node and send it to monitor
        df = pandas_node.to_arrow()
        return df



import json
import logging
import time
import urllib
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import attrs
import numpy
import pandas
import pendulum
from google.protobuf import json_format
from pyspark.sql import dataframe as pyspark_dataframe
from typeguard import typechecked

from tecton._internals import athena_api
from tecton._internals import display
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals import query_helper
from tecton._internals import querytree_api
from tecton._internals import sdk_decorators
from tecton._internals import snowflake_api
from tecton._internals import utils as internal_utils
from tecton._internals import validations_api
from tecton.framework import base_tecton_object
from tecton.framework import configs
from tecton.framework import data_frame as tecton_dataframe
from tecton.framework import feature_view as framework_feature_view
from tecton.framework import repo_config
from tecton.framework.data_frame import GetFeaturesForEventsParams
from tecton_core import conf
from tecton_core import fco_container
from tecton_core import feature_definition_wrapper
from tecton_core import feature_set_config
from tecton_core import http
from tecton_core import id_helper
from tecton_core import query_consts
from tecton_core import request_context
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.compute_mode import offline_retrieval_compute_mode
from tecton_core.repo_file_handler import construct_fco_source_info
from tecton_proto.api.featureservice import feature_service__client_pb2 as feature_service__api_pb2
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.args import fco_args__client_pb2 as fco_args_pb2
from tecton_proto.args import feature_service__client_pb2 as feature_service__args_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


logger = logging.getLogger(__name__)


@attrs.define(eq=False)
class FeatureService(base_tecton_object.BaseTectonObject):
    """A Tecton Feature Service.

    In Tecton, a Feature Service exposes an API for accessing a set of FeatureViews.

    Once deployed in production, each model has one associated Feature Service that serves the model its features. A
    Feature Service contains a list of the Feature Views associated with a model. It also includes user-provided
    metadata such as name, description, and owner that Tecton uses to organize feature data.

    An example of Feature Service declaration

    .. code-block:: python

        from tecton import FeatureService, LoggingConfig
        # Import your feature views declared in your feature repo directory
        from feature_repo.features.feature_views import last_transaction_amount_sql, transaction_amount_is_high
        ...

        # Declare Feature Service
        fraud_detection_feature_service = FeatureService(
            name='fraud_detection_feature_service',
            description='A FeatureService providing features for a model that predicts if a transaction is fraudulent.',
            features=[
                last_transaction_amount_sql,
                transaction_amount_is_high,
                ...
            ]
            logging=LoggingConfig(
                sample_rate=0.5,
                log_effective_times=False,
            )
            tags={'release': 'staging'},
        )

    Attributes:
        info: A dataclass containing basic info about this Tecton Object.
    """

    # A Feature Service spec, i.e. a dataclass representation of the Tecton object that is used in most functional use
    # cases, e.g. constructing queries. Set only after the object has been validated. Remote objects, i.e. applied
    # objects fetched from the backend, are assumed valid.
    _spec: Optional[specs.FeatureServiceSpec] = attrs.field(repr=False)

    # A Tecton "args" proto. Only set if this object was defined locally, i.e. this object was not applied
    # and fetched from the Tecton backend.
    _args: Optional[feature_service__args_pb2.FeatureServiceArgs] = attrs.field(
        repr=False, on_setattr=attrs.setters.frozen
    )

    # The feature references that make up this Feature Service.
    _feature_references: Tuple[framework_feature_view.FeatureReference, ...] = attrs.field(
        on_setattr=attrs.setters.frozen
    )

    # The feature set config for this Feature Service. The feature set config is used for query construction and
    # represents a super set of the feature references in _feature_references because of indirect feature definition
    # dependencies. For example, _feature_references may contain a single ODFV, but _feature_set_config may represent
    # that ODFV plus a batch feature view input to that ODFV.
    _feature_set_config: Optional[feature_set_config.FeatureSetConfig] = attrs.field(repr=False)

    # TODO(TEC-19861): remove once schemas are mandatory everywhere
    _all_fv_schemas_populated: bool = attrs.field(repr=False, init=False, default=True)

    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        online_serving_enabled: bool = True,
        features: Optional[
            List[Union[framework_feature_view.FeatureReference, framework_feature_view.FeatureView]]
        ] = None,
        logging: Optional[configs.LoggingConfig] = None,
        on_demand_environment: Optional[str] = None,
        options: Optional[Dict[str, str]] = None,
        enable_online_caching: bool = False,
    ):
        """
        Instantiates a new FeatureService.

        :param name: A unique name for the Feature Service.
        :param description: A human-readable description.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be set to False
            via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param online_serving_enabled: (Optional, default True) If True, users can send realtime requests
            to this FeatureService, and only FeatureViews with online materialization enabled can be added
            to this FeatureService.
        :param features: The list of FeatureView or FeatureReference that this FeatureService will serve.
        :param logging: A configuration for logging feature requests sent to this Feature Service.
        :param on_demand_environment: The environment in which all the on demand feature views for this feature service should be executed.
            Defaults to `None`, which means the on demand feature views are executed in the same environment as the
            feature service, without any resource isolation. This may be preferred for low-latency feature services
            which do not have dependencies. Learn more about environments at
            https://docs.tecton.ai/docs/defining-features/feature-views/on-demand-feature-view/on-demand-feature-view-environments.
        :param options: Additional options to configure the Feature Service. Used for advanced use cases and beta features.
        :param enable_online_caching: If True, the feature server will read and write feature values to the online serving cache for feature views
            that have caching enabled (have cache_config set).
        """
        if (
            on_demand_environment is None
            and repo_config.get_feature_service_defaults().on_demand_environment is not None
        ):
            on_demand_environment = repo_config.get_feature_service_defaults().on_demand_environment

        feature_references = []
        for feature in features:
            if isinstance(feature, framework_feature_view.FeatureReference):
                feature_references.append(feature)
            elif isinstance(feature, framework_feature_view.FeatureView):
                feature_references.append(framework_feature_view.FeatureReference(feature_definition=feature))
            else:
                msg = f"Object in FeatureService.features with an invalid type: {type(feature)}. Should be of type FeatureReference or Feature View."

        feature_references_args = [_feature_reference_to_feature_reference_args(ref) for ref in feature_references]

        args = feature_service__args_pb2.FeatureServiceArgs(
            feature_service_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            prevent_destroy=prevent_destroy,
            online_serving_enabled=online_serving_enabled,
            feature_references=feature_references_args,
            version=feature_definition_wrapper.FrameworkVersion.FWV5.value,
            logging=logging._to_proto() if logging else None,
            realtime_environment=on_demand_environment,
            options=options,
            enable_online_caching=enable_online_caching,
        )
        info = base_tecton_object.TectonObjectInfo.from_args_proto(args.info, args.feature_service_id)
        source_info = construct_fco_source_info(args.feature_service_id)
        self.__attrs_init__(
            info=info,
            spec=None,
            args=args,
            source_info=source_info,
            feature_references=feature_references,
            feature_set_config=None,
        )
        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._populate_spec()
        # TODO(TEC-19861): Remove once schemas are mandatory everywhere
        # Certain methods like `get_features_in_range` rely on the schemas being populated, and will raise an error if
        # it is not. They rely on the `_schema_populated` parameter, so that they don't have to re-compute this every
        # time.
        if any(not fv._schema_populated for fv in self.feature_views):
            self._all_fv_schemas_populated = False

        base_tecton_object._register_local_object(self)

    @classmethod
    @typechecked
    def _from_spec(cls, spec: specs.FeatureServiceSpec, fco_container: fco_container.FcoContainer) -> "FeatureService":
        """Create a Feature Service from directly from a spec. Specs are assumed valid and will not be re-validated."""
        feature_set_config_ = feature_set_config.FeatureSetConfig.from_feature_service_spec(spec, fco_container)
        info = base_tecton_object.TectonObjectInfo.from_spec(spec)

        feature_references = []
        for feature_set_item in spec.feature_set_items:
            fv_spec = fco_container.get_by_id(feature_set_item.feature_view_id)
            fv = framework_feature_view.feature_view_from_spec(fv_spec, fco_container)
            join_key_mapping = {
                jkm.feature_view_column_name: jkm.spine_column_name for jkm in feature_set_item.join_key_mappings
            }
            ref = framework_feature_view.FeatureReference(
                feature_definition=fv,
                namespace=feature_set_item.namespace,
                features=feature_set_item.feature_columns,
                override_join_keys=join_key_mapping,
            )
            feature_references.append(ref)

        obj = cls.__new__(cls)  # Instantiate the object. Does not call init.
        obj.__attrs_init__(
            info=info,
            spec=spec,
            args=None,
            source_info=None,
            feature_references=tuple(feature_references),
            feature_set_config=feature_set_config_,
        )

        return obj

    @sdk_decorators.assert_local_object
    def _build_args(self) -> fco_args_pb2.FcoArgs:
        return fco_args_pb2.FcoArgs(feature_service=self._args)

    def _build_fco_validation_args(self) -> validator_pb2.FcoValidationArgs:
        if self.info._is_local_object:
            return validator_pb2.FcoValidationArgs(
                feature_service=validator_pb2.FeatureServiceValidationArgs(
                    args=self._args,
                )
            )
        else:
            return self._spec.validation_args

    def _get_dependent_objects(self, include_indirect_dependencies: bool) -> List[base_tecton_object.BaseTectonObject]:
        dependent_objects = []
        for fv in self._feature_definitions:
            if include_indirect_dependencies:
                dependent_objects.extend([*fv._get_dependent_objects(include_indirect_dependencies=True), fv])
            else:
                dependent_objects.append(fv)

        # Dedupe by ID.
        return list({fco_obj.id: fco_obj for fco_obj in dependent_objects}.values())

    def _validate(self, indentation_level: int = 0) -> None:
        validations_api.run_backend_validation_and_assert_valid(
            self,
            validator_pb2.ValidationRequest(
                validation_args=[
                    dependent_obj._build_fco_validation_args()
                    for dependent_obj in self._get_dependent_objects(include_indirect_dependencies=True)
                ]
                + [self._build_fco_validation_args()],
            ),
            indentation_level,
        )

    def _populate_spec(self) -> None:
        dependent_specs = []
        for feature_definition in self._feature_definitions:
            dependent_specs.extend([*feature_definition._get_dependent_specs(), feature_definition._spec])

        supplement = specs.FeatureServiceSpecArgsSupplement(
            ids_to_feature_views={spec.id: spec for spec in dependent_specs if isinstance(spec, specs.FeatureViewSpec)}
        )
        fs_spec = specs.FeatureServiceSpec.from_args_proto(self._args, supplement)

        fco_container_specs = [fs_spec, *dependent_specs]
        fco_container_ = fco_container.FcoContainer.from_specs(specs=fco_container_specs, root_ids=[fs_spec.id])
        self._feature_set_config = feature_set_config.FeatureSetConfig.from_feature_service_spec(
            fs_spec, fco_container_
        )
        self._spec = fs_spec

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    def summary(self) -> display.Displayable:
        """Displays a human readable summary of this Feature View."""
        request = metadata_service_pb2.GetFeatureServiceSummaryRequest(
            feature_service_id=self._spec.id_proto, workspace=self._spec.workspace
        )
        response = metadata_service.instance().GetFeatureServiceSummary(request)

        items_map = {}
        summary_items = []
        for item in response.general_items:
            items_map[item.key] = item
            if item.display_name:
                summary_items.append((item.display_name, item.multi_values if item.multi_values else item.value))

        if not internal_utils.is_live_workspace(self.workspace):
            return display.Displayable.from_properties(items=summary_items)

        api_service = conf.get_or_raise("API_SERVICE")
        if "localhost" in api_service and "ingress" in api_service:
            return display.Displayable.from_properties(items=summary_items)

        if "curlEndpoint" in items_map and "curlParamsJson" in items_map:
            service_url = urllib.parse.urljoin(api_service, items_map["curlEndpoint"].value)
            curl_params_json = json.dumps(json.loads(items_map["curlParamsJson"].value), indent=2)
            curl_header = items_map["curlHeader"].value
            curl_str = "curl -X POST " + service_url + '\\\n-H "' + curl_header + "\"-d\\\n'" + curl_params_json + "'"
            summary_items.append(("Example cURL", curl_str))
        return display.Displayable.from_properties(items=summary_items)

    @property
    @sdk_decorators.sdk_public_method
    def features(self) -> List[framework_feature_view.FeatureReference]:
        """Returns the list of feature references included in this Feature Service.

        FeatureReferences are references to Feature Views/Tables that may select a subset of features, override the
        Feature View/Table namespace, or re-map join-keys.
        """
        return list(self._feature_references)

    @property
    @sdk_decorators.sdk_public_method
    def feature_views(self) -> Set[framework_feature_view.FeatureView]:
        """Returns the set of Feature Views directly depended on by this Feature Service.

        A single Feature View may be included multiple times in a Feature Service under different namespaces. See the
        FeatureReference documentation. This method dedupes those Feature Views.
        """
        # Dedupe by ids.
        # TODO(jake): Tecton objects should probably be hashable based on their id.
        return set({ref.feature_definition.id: ref.feature_definition for ref in self._feature_references}.values())

    @sdk_decorators.sdk_public_method
    def get_feature_columns(self) -> List[str]:
        """Returns the list of all feature columns included in this feature service."""
        self._validate_all_schemas_populated("get_feature_columns")
        all_features = self._feature_set_config.features
        return [feature for feature in all_features if not feature.startswith(query_consts.udf_internal())]

    def _validate_all_schemas_populated(self, method_name):
        if not self._all_fv_schemas_populated:
            raise errors.DEPENDENT_FEATURE_VIEWS_REQUIRE_SCHEMAS(
                method_name, self.name, [fv.name for fv in self.feature_views if not fv._schema_populated]
            )

    # TODO(follow-up PR): Clean up this method. (Holding off to minimize backport.)
    @property
    def _feature_definitions(self) -> Set[framework_feature_view.FeatureView]:
        """Returns the set of unique Feature Definitions directly depended on by this Feature Service.

        A single Feature Definition may be included multiple times in a Feature Service under different namespaces.
        This method dedupes those.
        """
        return {ref.feature_definition for ref in self._feature_references}

    def _check_can_query_from_source(self, from_source: Optional[bool]) -> framework_feature_view.QuerySources:
        all_query_sources = framework_feature_view.QuerySources()

        for fv in self._feature_definitions:
            sources = fv._check_can_query_from_source(from_source)
            all_query_sources += sources

        return all_query_sources

    @sdk_decorators.sdk_public_method
    def get_features_for_events(
        self,
        events: Union[pyspark_dataframe.DataFrame, pandas.DataFrame, tecton_dataframe.TectonDataFrame, str],
        timestamp_key: Optional[str] = None,
        from_source: Optional[bool] = None,
        compute_mode: Optional[Union[ComputeMode, str]] = None,
    ) -> tecton_dataframe.TectonDataFrame:
        """Fetch a :class:`TectonDataFrame` of feature values from this FeatureService.

        This method will return feature values for each row provided in the spine DataFrame. The feature values
        returned by this method will respect the timestamp provided in the timestamp column of the spine Data Frame.

        By default (i.e. ``from_source=None``), this method fetches feature values from the Offline Store for Feature
        Views that have offline materialization enabled and otherwise computes feature values on the fly from raw data.

        :param events: A dataframe of possible join keys, request data keys, and timestamps that specify which feature values to fetch.To distinguish
            between the event dataframe columns and feature columns, feature columns are labeled as `feature_view_name__feature_name`
            in the returned DataFrame.
        :type events: Union[pyspark.sql.DataFrame, pandas.DataFrame, TectonDataFrame]
        :param timestamp_key: Name of the time column in the spine DataFrame.
            This method will fetch the latest features computed before the specified timestamps in this column.
            Not applicable if the FeatureService stricly contains OnDemandFeatureViews with no feature view dependencies.
        :type timestamp_key: str
        :param from_source: Whether feature values should be recomputed from the original data source. If ``None``,
            feature values will be fetched from the Offline Store for Feature Views that have offline materialization
            enabled and otherwise computes feature values on the fly from raw data. Use ``from_source=True`` to force
            computing from raw data and ``from_source=False`` to error if any Feature Views are not materialized.
            Defaults to None.
        :type from_source: bool
        :param compute_mode: Compute mode to use to produce the data frame.

        Examples:
            A FeatureService :py:mod:`fs` that contains a BatchFeatureView and StreamFeatureView with join keys :py:mod:`user_id` and :py:mod:`ad_id`.

            1) :py:mod:`fs.get_features_for_events(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch features from the offline store for users 1, 2, and 3 for the specified timestamps and ad ids in the spine.

            2) :py:mod:`fv.get_features_for_events(spine, timestamp_key='date_1')` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'],
            'date_1': [datetime(...), ...], 'date_2': [datetime(...), ...]})`
            Fetch features from the offline store for users 1, 2, and 3 for the ad ids and specified timestamps in the :py:mod:`date_1` column in the spine.

            A FeatureService :py:mod:`fs_on_demand` that contains only OnDemandFeatureViews and expects request time data for the key :py:mod:`amount`.

            | The request time data struct is defined in the feature definition as such:
            | `request_schema = StructType()`
            | `request_schema.add(StructField('amount', DoubleType()))`
            | `transaction_request = RequestDataSource(request_schema=request_schema)`

            1) :py:mod:`fs_on_demand.get_features_for_events(spine)` where :py:mod:`spine=pandas.Dataframe({'amount': [30, 50, 10000]})`
            Fetch features from the offline store with request data inputs 30, 50, and 10000.

            A FeatureService :py:mod:`fs_all` that contains feature views of all types with join key 'user_id' and expects request time data for the key :py:mod:`amount`.

            1) :py:mod:`fs_all.get_features_for_events(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'amount': [30, 50, 10000], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch features from the offline store for users 1, 2, and 3 for the specified timestamps and request data inputs in the spine.

         :return: A Tecton :class:`TectonDataFrame`.
        """
        self._validate_all_schemas_populated("get_features_for_events")
        feature_view_text = internal_utils.plural(len(self._feature_definitions), "Feature View", "Feature Views")
        logger.info(f"Computing historical feature values for {len(self._feature_definitions)} {feature_view_text}.")

        sources = self._check_can_query_from_source(from_source)
        sources.display()

        compute_mode = offline_retrieval_compute_mode(compute_mode)

        if compute_mode == ComputeMode.SNOWFLAKE or compute_mode == ComputeMode.ATHENA:
            raise errors.GET_FEATURES_FOR_EVENTS_UNSUPPORTED

        df = querytree_api.get_historical_features_for_feature_service(
            dialect=compute_mode.default_dialect(),
            compute_mode=compute_mode,
            feature_service_spec=self._spec,
            feature_set_config=self._feature_set_config,
            spine=events,
            timestamp_key=timestamp_key,
            from_source=from_source,
        )
        df._requestContext = GetFeaturesForEventsParams(
            self._spec,
            events,
            timestamp_key,
            from_source,
            compute_mode,
        )
        return df

    @sdk_decorators.sdk_public_method
    def get_historical_features(
        self,
        spine: Union[pyspark_dataframe.DataFrame, pandas.DataFrame, tecton_dataframe.TectonDataFrame, str],
        timestamp_key: Optional[str] = None,
        include_feature_view_timestamp_columns: bool = False,
        from_source: Optional[bool] = None,
        compute_mode: Optional[Union[ComputeMode, str]] = None,
    ) -> tecton_dataframe.TectonDataFrame:
        """Fetch a :class:`TectonDataFrame` of feature values from this FeatureService.

        This method will return feature values for each row provided in the spine DataFrame. The feature values
        returned by this method will respect the timestamp provided in the timestamp column of the spine DataFrame.

        By default (i.e. ``from_source=None``), this method fetches feature values from the Offline Store for Feature
        Views that have offline materialization enabled and otherwise computes feature values on the fly from raw data.

        :param spine: A dataframe of possible join keys, request data keys, and timestamps that specify which feature values to fetch.To distinguish
            between spine columns and feature columns, feature columns are labeled as `feature_view_name__feature_name`
            in the returned DataFrame.
        :type spine: Union[pyspark.sql.DataFrame, pandas.DataFrame, TectonDataFrame]
        :param timestamp_key: Name of the time column in the spine DataFrame.
            This method will fetch the latest features computed before the specified timestamps in this column.
            Not applicable if the FeatureService stricly contains OnDemandFeatureViews with no feature view dependencies.
        :type timestamp_key: str
        :param include_feature_view_timestamp_columns: Whether to include timestamp columns for each FeatureView in the FeatureService. Default is False.
        :type include_feature_view_timestamp_columns: bool
        :param from_source: Whether feature values should be recomputed from the original data source. If ``None``,
            feature values will be fetched from the Offline Store for Feature Views that have offline materialization
            enabled and otherwise computes feature values on the fly from raw data. Use ``from_source=True`` to force
            computing from raw data and ``from_source=False`` to error if any Feature Views are not materialized.
            Defaults to None.
        :type from_source: bool
        :param compute_mode: Compute mode to use to produce the data frame.

        Examples:
            A FeatureService :py:mod:`fs` that contains a BatchFeatureView and StreamFeatureView with join keys :py:mod:`user_id` and :py:mod:`ad_id`.

            1) :py:mod:`fs.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps and ad ids in the spine.

            2) :py:mod:`fv.get_historical_features(spine, timestamp_key='date_1')` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'],
            'date_1': [datetime(...), ...], 'date_2': [datetime(...), ...]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the ad ids and specified timestamps in the :py:mod:`date_1` column in the spine.

            A FeatureService :py:mod:`fs_on_demand` that contains only OnDemandFeatureViews and expects request time data for the key :py:mod:`amount`.

            | The request time data is defined in the feature definition as such:
            | `request_schema = StructType()`
            | `request_schema.add(StructField('amount', DoubleType()))`
            | `transaction_request = RequestDataSource(request_schema=request_schema)`

            1) :py:mod:`fs_on_demand.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'amount': [30, 50, 10000]})`
            Fetch historical features from the offline store with request data inputs 30, 50, and 10000.

            A FeatureService :py:mod:`fs_all` that contains feature views of all types with join key 'user_id' and expects request time data for the key :py:mod:`amount`.

            1) :py:mod:`fs_all.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'amount': [30, 50, 10000], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps and request data inputs in the spine.

         :return: A Tecton :class:`TectonDataFrame`.
        """
        self._validate_all_schemas_populated("get_historical_features")

        feature_view_text = internal_utils.plural(len(self._feature_definitions), "Feature View", "Feature Views")
        logger.info(f"Computing historical feature values for {len(self._feature_definitions)} {feature_view_text}.")
        sources = self._check_can_query_from_source(from_source)
        sources.display()

        compute_mode = offline_retrieval_compute_mode(compute_mode)

        has_snowpark_transformation = snowflake_api.has_snowpark_transformation(
            [feature_view._feature_definition for feature_view in self._feature_definitions]
        )

        if compute_mode == ComputeMode.SNOWFLAKE and (
            conf.get_bool("USE_DEPRECATED_SNOWFLAKE_RETRIEVAL") or has_snowpark_transformation
        ):
            # Snowflake retrieval is now migrated to QT and this code path is deprecated.
            if has_snowpark_transformation:
                logger.warning(
                    "At least one Feature View in this Feature Service uses a snowpark transformation. Snowpark transformations are deprecated in versions >=0.8. Consider using snowflake_sql transformations instead."
                )
            return snowflake_api.get_historical_features(
                spine=spine,
                timestamp_key=timestamp_key,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                from_source=from_source,
                feature_set_config=self._feature_set_config,
            )

        if compute_mode == ComputeMode.ATHENA and conf.get_bool("USE_DEPRECATED_ATHENA_RETRIEVAL"):
            if not internal_utils.is_live_workspace(self.workspace):
                raise errors.ATHENA_COMPUTE_ONLY_SUPPORTED_IN_LIVE_WORKSPACE
            return athena_api.get_historical_features(
                spine=spine,
                timestamp_key=timestamp_key,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                from_source=from_source,
                feature_set_config=self._feature_set_config,
            )

        logger.warning(errors.GET_HISTORICAL_FEATURES_DEPRECATED_SPINE)

        return querytree_api.get_historical_features_for_feature_service(
            dialect=compute_mode.default_dialect(),
            compute_mode=compute_mode,
            feature_service_spec=self._spec,
            feature_set_config=self._feature_set_config,
            spine=spine,
            timestamp_key=timestamp_key,
            from_source=from_source,
        )

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    @typechecked
    def get_online_features(
        self,
        join_keys: Optional[Mapping[str, Union[int, numpy.int_, str, bytes]]] = None,
        include_join_keys_in_response: bool = False,
        request_data: Optional[Mapping[str, Union[int, numpy.int_, str, bytes, float]]] = None,
    ) -> tecton_dataframe.FeatureVector:
        """
        Returns a single Tecton :class:`tecton.FeatureVector` from the Online Store.
        At least one of join_keys or request_data is required.

        :param join_keys: Join keys of the enclosed FeatureViews.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.
        :param request_data: Dictionary of request context values. Only applicable when the FeatureService contains OnDemandFeatureViews.

        Examples:
            A FeatureService :py:mod:`fs` that contains a BatchFeatureView and StreamFeatureView with join keys :py:mod:`user_id` and :py:mod:`ad_id`.

            1) :py:mod:`fs.get_online_features(join_keys={'user_id': 1, 'ad_id': 'c234'})`
            Fetch the latest features from the online store for user 1 and ad 'c234'.

            2) :py:mod:`fv.get_online_features(join_keys={'user_id': 1, 'ad_id': 'c234'}, include_join_keys_in_respone=True)`
            Fetch the latest features from the online store for user 1 and ad id 'c234'. Include the join key information (user_id=1, ad_id='c234') in the returned FeatureVector.

            A FeatureService :py:mod:`fs_on_demand` that contains only OnDemandFeatureViews and expects request time data for key :py:mod:`amount`.

            | The request time data is defined in the feature definition as such:
            | `request_schema = StructType()`
            | `request_schema.add(StructField('amount', DoubleType()))`
            | `transaction_request = RequestDataSource(request_schema=request_schema)`

            1) :py:mod:`fs_on_demand.get_online_features(request_data={'amount': 30})`
            Fetch the latest features from the online store with amount = 30.

            A FeatureService :py:mod:`fs_all` that contains feature views of all types with join key :py:mod:`user_id` and expects request time data for key :py:mod:`amount`.

            1) :py:mod:`fs_all.get_online_features(join_keys={'user_id': 1}, request_data={'amount': 30})`
            Fetch the latest features from the online store for user 1 with amount = 30.

        :return: A :class:`tecton.FeatureVector` of the results.
        """
        # Default to empty dicts.
        join_keys = join_keys or {}
        request_data = request_data or {}

        if not internal_utils.is_live_workspace(self.workspace):
            msg = "get_online_features"
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE(msg)

        if not self._spec.online_serving_enabled:
            msg = "get_online_features"
            raise errors.UNSUPPORTED_OPERATION(msg, "online_serving_enabled was not defined for this Feature Service.")
        if not join_keys and not request_data:
            raise errors.FS_GET_ONLINE_FEATURES_REQUIRED_ARGS

        fs_contains_non_odfvs = not all(fd.is_realtime for fd in self._feature_set_config.feature_definitions)
        if fs_contains_non_odfvs and not join_keys:
            raise errors.GET_ONLINE_FEATURES_FS_JOIN_KEYS

        request_context = self._request_context
        required_request_context_keys = list(request_context.schema.keys())
        if required_request_context_keys and not request_data:
            raise errors.GET_ONLINE_FEATURES_FS_NO_REQUEST_DATA(required_request_context_keys)
        internal_utils.validate_request_data(request_data, required_request_context_keys)

        return query_helper._QueryHelper(self.workspace, feature_service_name=self.name).get_feature_vector(
            join_keys,
            include_join_keys_in_response,
            request_data,
            request_context,
        )

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    @typechecked
    def query_features(
        self, join_keys: Mapping[str, Union[numpy.int_, int, str, bytes]]
    ) -> tecton_dataframe.TectonDataFrame:
        """
        [Advanced Feature] Queries the FeatureService with a partial set of join_keys defined in the ``online_serving_index``
        of the included FeatureViews. Returns a Tecton :class:`TectonDataFrame` of all matched records.

        :param join_keys: Query join keys, i.e., a union of join keys in the ``online_serving_index`` of all
            enclosed FeatureViews.
        :return: A Tecton :class:`TectonDataFrame`
        """
        if not internal_utils.is_live_workspace(self.workspace):
            msg = "query_features"
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE(msg)

        if not self._spec.online_serving_enabled:
            msg = "query_features"
            raise errors.UNSUPPORTED_OPERATION(msg, "online_serving_enabled was not defined for this Feature Service.")

        return query_helper._QueryHelper(self.workspace, feature_service_name=self.name).query_features(join_keys)

    @property
    def _request_context(self) -> request_context.RequestContext:
        merged_context = request_context.RequestContext({})
        for fv in self._feature_definitions:
            if isinstance(fv, framework_feature_view.RealtimeFeatureView):
                merged_context.merge(fv._request_context)
        return merged_context

    @sdk_decorators.assert_remote_object
    def _wait_until_ready(
        self, timeout: Optional[pendulum.Duration] = None, wait_for_materialization=True, verbose=False
    ):
        """Blocks until the service is ready to serve real-time requests.

        The FeatureService is considered ready once every FeatureView that has been added to it
        has had at least once successful materialization run.

        :param timeout: The timeout to wait. Defaults to 15 minutes.
        :param wait_for_materialization: If False, does not wait for batch materialization to complete.
        """
        if not internal_utils.is_live_workspace(self.workspace):
            msg = "_wait_until_ready"
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE(msg)

        if timeout is None:
            timeout = pendulum.Duration(minutes=15)

        deadline = pendulum.now() + timeout

        has_been_not_ready = False
        while True:
            request = feature_service__api_pb2.GetFeatureServiceStateRequest()
            request.feature_service_locator.feature_service_name = self.name
            request.feature_service_locator.workspace_name = self.workspace
            http_response = http.session().post(
                urllib.parse.urljoin(
                    conf.get_or_raise("FEATURE_SERVICE") + "/", "v1/feature-service/get-feature-service-state"
                ),
                data=json_format.MessageToJson(request),
                headers=query_helper._QueryHelper(self.workspace, feature_service_name=self.name)._prepare_headers(),
            )
            details = http_response.json()
            if http_response.status_code == 404:
                # FeatureService is not ready to serve
                if verbose:
                    logger.info(f" Waiting for FeatureService to be ready to serve ({details['message']})")
                else:
                    logger.info(" Waiting for FeatureService to be ready to serve")
            elif http_response.status_code == 200:
                # FeatureService is ready
                if verbose:
                    logger.info(f"wait_until_ready: Ready! Response={http_response.text}")
                else:
                    logger.info("wait_until_ready: Ready!")
                # Extra wait time due to different FS hosts being potentially out-of-sync in picking up the latest state
                if has_been_not_ready:
                    time.sleep(20)
                return
            else:
                http_response.raise_for_status()
                return
            if pendulum.now() > deadline:
                logger.info(f"wait_until_ready: Response={http_response.text}")
                raise TimeoutError()
            has_been_not_ready = True
            time.sleep(10)
            continue


def _feature_reference_to_feature_reference_args(
    feature_reference: framework_feature_view.FeatureReference,
) -> feature_service__args_pb2.FeatureReference:
    if feature_reference.override_join_keys:
        override_join_keys = [
            feature_service__args_pb2.ColumnPair(feature_column=fv_key, spine_column=spine_key)
            for fv_key, spine_key in sorted(feature_reference.override_join_keys.items())
        ]
    else:
        override_join_keys = None

    return feature_service__args_pb2.FeatureReference(
        feature_view_id=feature_reference.feature_definition._id_proto,
        namespace=feature_reference.namespace,
        override_join_keys=override_join_keys,
        features=feature_reference.features if feature_reference.features else None,
    )



from datetime import datetime
from unittest import TestCase
from unittest import mock

from google.protobuf.timestamp_pb2 import Timestamp
from pytz import UTC

from tecton.cli.plan import PlanListItem
from tecton.cli.plan import PlanSummary
from tecton.cli.plan import get_plans_list_items
from tecton_proto.auth import principal__client_pb2 as principal_pb2
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


def make_principle_basic(first_name="john", last_name="smith"):
    return principal_pb2.PrincipalBasic(
        user=principal_pb2.UserBasic(
            okta_id="okta-id",
            first_name=first_name,
            last_name=last_name,
            login_email=f"{first_name}.{last_name}@tecon.ai",
        )
    )


class BaseTestCase(TestCase):
    def mockPatch(self, *args, **kwargs):
        patcher = mock.patch(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def setUp(self) -> None:
        self.mock_metadata_service = mock.MagicMock()
        self.mockPatch("tecton._internals.metadata_service.instance", return_value=self.mock_metadata_service)


class TestListPlan(BaseTestCase):
    maxDiff = 10000

    def test_plan_list_item_from_proto(self):
        test_date = datetime(year=2024, month=1, day=1)
        pb_test_date = Timestamp()
        pb_test_date.FromDatetime(test_date)
        self.assertEqual(pb_test_date.ToDatetime(), test_date)
        plan_list_item = PlanListItem.from_proto(
            state_update_pb2.StateUpdateEntry(
                commit_id="fake_commit",
                applied_at=pb_test_date,
                applied_by="me",
                applied_by_principal=make_principle_basic(),
                workspace="here",
                sdk_version="0.0.0",
            )
        )
        self.assertEqual(
            plan_list_item,
            PlanListItem(
                plan_id="fake_commit",
                applied_by="john.smith@tecon.ai",
                applied_at=datetime(year=2024, month=1, day=1, tzinfo=UTC),
                workspace="here",
                sdk_version="0.0.0",
            ),
        )

    def test_plan_list_items_from_proto_empty(self):
        plan_list_item = PlanListItem.from_proto(state_update_pb2.StateUpdateEntry())
        self.assertEqual(
            plan_list_item, PlanListItem(plan_id="", applied_by="", applied_at=None, workspace="", sdk_version="")
        )

    def test_get_plans_list_items(self):
        self.mock_metadata_service.GetStateUpdateLog.return_value = metadata_service_pb2.GetStateUpdateLogResponse(
            entries=[
                state_update_pb2.StateUpdateEntry(
                    commit_id="fake-commit", applied_by="me", workspace="my_workspace", sdk_version="0.0.0"
                ),
                state_update_pb2.StateUpdateEntry(
                    commit_id="fake-commit2", applied_by="you", workspace="my_workspace", sdk_version="0.0.1"
                ),
            ]
        )
        items = get_plans_list_items("my_workspace", 100)
        self.mock_metadata_service.GetStateUpdateLog.assert_called_with(
            metadata_service_pb2.GetStateUpdateLogRequest(workspace="my_workspace", limit=100)
        )
        assert items == [
            PlanListItem(
                plan_id="fake-commit",
                applied_by="me",
                applied_at=None,
                workspace="my_workspace",
                sdk_version="0.0.0",
            ),
            PlanListItem(
                plan_id="fake-commit2",
                applied_by="you",
                applied_at=None,
                workspace="my_workspace",
                sdk_version="0.0.1",
            ),
        ]

    def test_get_plans_list_items_empty(self):
        self.mock_metadata_service.GetStateUpdateLog.return_value = metadata_service_pb2.GetStateUpdateLogResponse()
        items = get_plans_list_items("my_workspace", 100)
        self.mock_metadata_service.GetStateUpdateLog.assert_called_with(
            metadata_service_pb2.GetStateUpdateLogRequest(workspace="my_workspace", limit=100)
        )
        assert items == []


class TestGetPlan(BaseTestCase):
    def test_plan_summary_from_proto_empty(self):
        plan = PlanSummary.from_proto(state_update_pb2.StateUpdatePlanSummary())
        self.assertEqual(
            plan,
            PlanSummary(
                applied=False,
                applied_by="",
                applied_at=None,
                created_by="",
                created_at=None,
                workspace="",
                sdk_version="",
            ),
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/resource__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(tecton_proto/auth/resource__client.proto\x12\x11tecton_proto.auth*\xd6\x04\n\x0cResourceType\x12\x1d\n\x19RESOURCE_TYPE_UNSPECIFIED\x10\x00\x12\x16\n\x12RESOURCE_TYPE_USER\x10\x01\x12\x1e\n\x1aRESOURCE_TYPE_ORGANIZATION\x10\x02\x12\x1b\n\x17RESOURCE_TYPE_WORKSPACE\x10\x03\x12!\n\x1dRESOURCE_TYPE_SERVICE_ACCOUNT\x10\x04\x12!\n\x1dRESOURCE_TYPE_PRINCIPAL_GROUP\x10\x05\x12\x18\n\x14RESOURCE_TYPE_SECRET\x10\x06\x12\x18\n\x14RESOURCE_TYPE_ACCESS\x10\x07\x12%\n!RESOURCE_TYPE_DEPLOYMENT_SETTINGS\x10\x08\x12\x19\n\x15RESOURCE_TYPE_DATASET\x10\t\x12\x1a\n\x16RESOURCE_TYPE_FEATURES\x10\n\x12\x18\n\x14RESOURCE_TYPE_CONFIG\x10\x0b\x12\x18\n\x14RESOURCE_TYPE_METRIC\x10\x0c\x12\x16\n\x12RESOURCE_TYPE_DATA\x10\r\x12\'\n#RESOURCE_TYPE_MATERIALIZATION_STATE\x10\x0e\x12\x19\n\x15RESOURCE_TYPE_MEMBERS\x10\x0f\x12.\n*RESOURCE_TYPE_PRINCIPAL_GROUPS_MEMBERSHIPS\x10\x10\x12\x1a\n\x16RESOURCE_TYPE_PLATFORM\x10\x11\x12\x1e\n\x1aRESOURCE_TYPE_SECRET_SCOPE\x10\x12\x42J\n\x0f\x63om.tecton.authP\x01Z5github.com/tecton-ai/tecton_proto/auth/resource_roles')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.resource__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authP\001Z5github.com/tecton-ai/tecton_proto/auth/resource_roles'
  _RESOURCETYPE._serialized_start=64
  _RESOURCETYPE._serialized_end=662
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import importlib
from collections import abc
from dataclasses import dataclass
from pathlib import Path
from types import MappingProxyType
from typing import Any
from typing import Callable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import Optional

import numpy
import pyarrow
import torch

from tecton_core import data_types
from tecton_core import schema


_LOAD_FN_NAME = "load_context"
_PRE_PROCESSOR_FN_NAME = "preprocessor"
_POST_PROCESSOR_FN_NAME = "postprocessor"
_RESERVED_MODEL_CONTEXT_KEY = "model"
_TECTON_MODEL_SPEC_NAME = "_tecton_custom_model"

_CONTEXT_PARAM = Mapping[str, Any]
_PRE_PROCESSOR_FN = Optional[Callable[[Mapping[str, numpy.ndarray], _CONTEXT_PARAM], Mapping[str, torch.Tensor]]]
_POST_PROCESSOR_FN = Optional[Callable[[torch.Tensor, _CONTEXT_PARAM], torch.Tensor]]

# TODO(jiadong): Implement this as a function to handle complex data type.
_TECTON_TYPE_TO_ARROW_TYPE: Mapping[data_types.DataType, pyarrow.DataType] = {
    data_types.Int32Type(): pyarrow.int32(),
    data_types.Int64Type(): pyarrow.int64(),
    data_types.Float32Type(): pyarrow.float32(),
    data_types.Float64Type(): pyarrow.float64(),
    data_types.StringType(): pyarrow.string(),
}


def _default_pytorch_preprocessor(input: Mapping[str, numpy.ndarray]) -> Mapping[str, torch.Tensor]:
    return {name: torch.tensor(array) for name, array in input.items()}


# TODO(jiadong): CustomModelContainer is pytorch based currently. It can be refactored to a structure that has a `BaseModelContainer` and framework specific model container becomes subclasses of it. The base class can have `load` function implmeneted which is framework agonistic and each subclass implements the `predict` function.
class CustomModelContainer:
    _data_dir: Path

    # Late init params
    _context: MutableMapping[str, Any]

    _preprocessor: _PRE_PROCESSOR_FN = None
    _postprocessor: _POST_PROCESSOR_FN = None

    def __init__(self, data_dir: str) -> None:
        self._data_dir = Path(data_dir)
        self._context = {}

    def load(self):
        model_module = self._load_model_module()

        _load_fn = getattr(model_module, _LOAD_FN_NAME, None)
        if not _load_fn:
            msg = f"`{_LOAD_FN_NAME}` function is missing in the model repo"
            raise AttributeError(msg)

        _load_fn(self._data_dir, self._context)

        model = self._context.get(_RESERVED_MODEL_CONTEXT_KEY, None)
        if not model:
            msg = "No 'model' found in the context. `load` function should initialize the model and put it in the context with 'model' key."
            raise ValueError(msg)

        self._preprocessor = getattr(model_module, _PRE_PROCESSOR_FN_NAME, None)
        self._postprocessor = getattr(model_module, _POST_PROCESSOR_FN_NAME, None)

    def _load_model_module(self):
        # TODO(jiadong): Read main python file name from ModelArtifactInfo instead of using hardcoded `model.py`
        spec = importlib.util.spec_from_file_location(_TECTON_MODEL_SPEC_NAME, self._data_dir / "model.py")
        model_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(model_module)
        return model_module

    def predict(self, input: pyarrow.RecordBatch) -> torch.Tensor:
        numpy_dict = {name: column.to_numpy() for name, column in zip(input.schema.names, input.columns)}
        if self._preprocessor:
            input = self._preprocessor(numpy_dict, self._context)
        else:
            input = _default_pytorch_preprocessor(numpy_dict)

        if not isinstance(input, abc.Mapping):
            msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]', but received '{type(input)}'"
            raise TypeError(msg)

        for k, v in input.items():
            if not isinstance(v, torch.Tensor):
                msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]' but detected '{k}' returned by preprocessor is an instance of type {type(v)}."
                raise TypeError(msg)

        model = self._context[_RESERVED_MODEL_CONTEXT_KEY]
        output = model(**input)

        if self._postprocessor:
            output = self._postprocessor(output, self._context)

            if not isinstance(output, torch.Tensor):
                msg = (
                    f"`postprocessor` needs to return a 'torch.Tensor' but detected an instance of type {type(output)}"
                )
                raise TypeError(msg)

        return output


# TODO(EMBED-117): Bisect input data if the inference OOM.
def custom_model_inference(
    batch: pyarrow.RecordBatch,
    *,
    model: CustomModelContainer,
    input_columns: List[schema.Column],
    output_column: schema.Column,
    cuda_device: str,
) -> pyarrow.RecordBatch:
    with torch.device(cuda_device):
        input_batch = batch.select([col.name for col in input_columns])
        output = model.predict(input_batch)
        output = output.cpu().numpy()
        output = pyarrow.array(list(output), type=_TECTON_TYPE_TO_ARROW_TYPE[output_column.dtype])

        batch_with_results = pyarrow.RecordBatch.from_arrays(
            [*batch.columns, output],
            schema=batch.schema.append(pyarrow.field(output_column.name, output.type)),
        )
    return batch_with_results


# TODO(jiadong): Implement custom model batch function.
def default_custom_batch_fn(
    batch: pyarrow.RecordBatch,
) -> Iterator[pyarrow.RecordBatch]:
    yield batch


@dataclass
class CustomModelInferenceFuncConfig:
    _model_dir: str
    _extra_kwargs: Mapping[str, Any]
    _cuda_device: str

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(
        cls,
        model_dir: str,
        cuda_device: str,
        input_columns: List[schema.Column],
        output_column: schema.Column,
    ) -> CustomModelInferenceFuncConfig:
        return cls(
            _model_dir=model_dir,
            _cuda_device=cuda_device,
            _extra_kwargs=MappingProxyType(
                {
                    "cuda_device": cuda_device,
                    "input_columns": input_columns,
                    "output_column": output_column,
                }
            ),
        )

    def load(self):
        model_container = CustomModelContainer(data_dir=self._model_dir)
        with torch.device(self._cuda_device):
            model_container.load()

        self._final_kwargs = MappingProxyType(dict(model=model_container, **self._extra_kwargs))
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs


@dataclass
class CustomModelBatchFuncConfig:
    # TODO(jiadong): Add necessary params to batch function config such as user-provided batch size.

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(cls) -> CustomModelBatchFuncConfig:
        return cls()

    def load(self):
        self._final_kwargs = MappingProxyType({})
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/aggregation_function__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6tecton_proto/common/aggregation_function__client.proto\x12\x13tecton_proto.common\"\x1b\n\x0bLastNParams\x12\x0c\n\x01n\x18\x01 \x01(\x05R\x01n\"\x1c\n\x0c\x46irstNParams\x12\x0c\n\x01n\x18\x01 \x01(\x05R\x01n\"9\n\x19\x41pproxCountDistinctParams\x12\x1c\n\tprecision\x18\x01 \x01(\x05R\tprecision\"V\n\x16\x41pproxPercentileParams\x12\x1e\n\npercentile\x18\x01 \x01(\x01R\npercentile\x12\x1c\n\tprecision\x18\x02 \x01(\x05R\tprecision\"\xe0\x02\n\x19\x41ggregationFunctionParams\x12\x39\n\x06last_n\x18\x01 \x01(\x0b\x32 .tecton_proto.common.LastNParamsH\x00R\x05lastN\x12<\n\x07\x66irst_n\x18\x02 \x01(\x0b\x32!.tecton_proto.common.FirstNParamsH\x00R\x06\x66irstN\x12\x64\n\x15\x61pprox_count_distinct\x18\x03 \x01(\x0b\x32..tecton_proto.common.ApproxCountDistinctParamsH\x00R\x13\x61pproxCountDistinct\x12Z\n\x11\x61pprox_percentile\x18\x04 \x01(\x0b\x32+.tecton_proto.common.ApproxPercentileParamsH\x00R\x10\x61pproxPercentileB\x08\n\x06params*\x99\x05\n\x13\x41ggregationFunction\x12 \n\x1c\x41GGREGATION_FUNCTION_UNKNOWN\x10\x00\x12\x1e\n\x1a\x41GGREGATION_FUNCTION_COUNT\x10\x01\x12\x1c\n\x18\x41GGREGATION_FUNCTION_SUM\x10\x02\x12\x1d\n\x19\x41GGREGATION_FUNCTION_MEAN\x10\x03\x12\x1d\n\x19\x41GGREGATION_FUNCTION_LAST\x10\x05\x12\x1c\n\x18\x41GGREGATION_FUNCTION_MIN\x10\x06\x12\x1c\n\x18\x41GGREGATION_FUNCTION_MAX\x10\x07\x12(\n$AGGREGATION_FUNCTION_LAST_DISTINCT_N\x10\t\x12!\n\x1d\x41GGREGATION_FUNCTION_VAR_SAMP\x10\n\x12 \n\x1c\x41GGREGATION_FUNCTION_VAR_POP\x10\x0b\x12$\n AGGREGATION_FUNCTION_STDDEV_SAMP\x10\x0c\x12#\n\x1f\x41GGREGATION_FUNCTION_STDDEV_POP\x10\r\x12,\n(AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N\x10\x0e\x12-\n)AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N\x10\x0f\x12)\n%AGGREGATION_FUNCTION_FIRST_DISTINCT_N\x10\x10\x12.\n*AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT\x10\x11\x12*\n&AGGREGATION_FUNCTION_APPROX_PERCENTILE\x10\x12\"\x04\x08\x04\x10\x04\"\x04\x08\x08\x10\x08\x42?\n\x11\x63om.tecton.commonP\x01Z(github.com/tecton-ai/tecton_proto/common')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.aggregation_function__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001Z(github.com/tecton-ai/tecton_proto/common'
  _AGGREGATIONFUNCTION._serialized_start=641
  _AGGREGATIONFUNCTION._serialized_end=1306
  _LASTNPARAMS._serialized_start=79
  _LASTNPARAMS._serialized_end=106
  _FIRSTNPARAMS._serialized_start=108
  _FIRSTNPARAMS._serialized_end=136
  _APPROXCOUNTDISTINCTPARAMS._serialized_start=138
  _APPROXCOUNTDISTINCTPARAMS._serialized_end=195
  _APPROXPERCENTILEPARAMS._serialized_start=197
  _APPROXPERCENTILEPARAMS._serialized_end=283
  _AGGREGATIONFUNCTIONPARAMS._serialized_start=286
  _AGGREGATIONFUNCTIONPARAMS._serialized_end=638
# @@protoc_insertion_point(module_scope)



import logging
import sys
import time
import urllib.parse
from pathlib import Path
from typing import Optional

import click

import tecton
from tecton import conf
from tecton import tecton_context
from tecton._internals import metadata_service
from tecton._internals import sdk_decorators
from tecton._internals.utils import cluster_url
from tecton.cli import access_control
from tecton.cli import api_key
from tecton.cli import auth
from tecton.cli import completion
from tecton.cli import engine
from tecton.cli import environment
from tecton.cli import materialization
from tecton.cli import plan
from tecton.cli import printer
from tecton.cli import repo
from tecton.cli import repo_config
from tecton.cli import secrets
from tecton.cli import service_account
from tecton.cli import test
from tecton.cli import user
from tecton.cli import workspace
from tecton.cli.command import TectonGroup
from tecton.cli.engine import dump_local_state
from tecton.cli.repo_utils import get_tecton_objects_skip_validation
from tecton.cli.workspace_utils import WorkspaceType


CONTEXT_SETTINGS = {
    "max_content_width": 160,
    "help_option_names": ["-h", "--help"],
}


@click.group(name="tecton", context_settings=CONTEXT_SETTINGS, cls=TectonGroup)
@click.option("--verbose", "-v", is_flag=True, help="Increase verbosity level to print more information.")
def cli(verbose: bool = False):
    """Tecton command-line tool."""
    sdk_decorators.disable_sdk_public_method_decorator()
    conf.enable_save_tecton_configs()

    logging_level = logging.DEBUG if verbose else logging.WARNING
    logging.basicConfig(
        level=logging_level,
        stream=sys.stderr,
        format="%(levelname)s(%(name)s): %(message)s",
    )

    # add cwd to path
    sys.path.append("")


cli.add_command(auth.login)
cli.add_command(auth.logout)
cli.add_command(auth.whoami)
cli.add_command(access_control.access_control)
cli.add_command(api_key.api_key)
cli.add_command(completion.completion)
cli.add_command(engine.apply)
cli.add_command(engine.destroy)
cli.add_command(engine.upgrade)
cli.add_command(environment.environment)
cli.add_command(materialization.materialization_status)
cli.add_command(materialization.freshness)
cli.add_command(repo.init)
cli.add_command(repo.restore)
cli.add_command(secrets.secrets)
cli.add_command(service_account.service_account)
cli.add_command(test.test)
cli.add_command(user.user)
cli.add_command(plan.plan)
cli.add_command(workspace.workspace)
cli.add_command(repo_config.repo_config_group)


@cli.command(requires_auth=False)
def version():
    """Print CLI version."""
    tecton.version.summary()


@cli.command(hidden=True)
@click.option(
    "--config",
    help="Path to the repo config yaml file. Defaults to the repo.yaml file at the Tecton repo root.",
    default=None,
    type=click.Path(exists=True, dir_okay=False, path_type=Path, readable=True),
)
def dump(config: Optional[Path]) -> None:
    """Print the serialization format for object definitions in this repository.

    Useful for debugging issues related to serializing the Tecton repository to be sent to the Tecton backend.
    """
    # TODO(jake): get_tecton_objects prints out to stdout, which breaks piping the `tecton dump` output. Start printing
    # to stderr.
    top_level_objects, _, _, _ = get_tecton_objects_skip_validation(config)
    dump_local_state(top_level_objects)


@cli.command(uses_workspace=True, hidden=True)
@click.option("--limit", default=10, type=int, help="Number of log entries to return.")
def log(limit):
    """View log of applied plans in workspace."""
    # TODO: Remove before tecton version 1.1.0
    printer.safe_print()
    plan.list_impl(limit)


@cli.command()
@click.option(
    "--workspace",
    default=None,
    type=WorkspaceType(),
    help="Workspace selected in Tecton Web UI. Defaults to the current selected workspace.",
)
@click.option(
    "--print/--no-print", "-p", "print_", default=False, help="Print URL instead of automatically launching a browser."
)
def web(workspace, print_) -> None:
    """Opens Tecton UI in a browser."""
    workspace_name = workspace if workspace else tecton_context.get_current_workspace()
    if workspace_name:
        web_url = urllib.parse.urljoin(cluster_url(), f"app/repo/{workspace_name}/")
    else:
        web_url = urllib.parse.urljoin(cluster_url(), "app")

    if print_:
        printer.safe_print(f"Web URL: {web_url}")
    else:
        printer.safe_print(f"Opening {web_url}")
        # Sleep before opening the browser to improve the UX and make it less jarring.
        time.sleep(1)
        click.launch(web_url)


def main():
    try:
        cli()
    finally:
        metadata_service.close_instance()


if __name__ == "__main__":
    main()
